#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Functions for working with Unicode characters, codepoints, and surrogate pairs.
"""
import unicodedata

INVISIBLE_CHARS = (
    "͏",
    "ᅟ",
    "ᅠ",
    "឴",
    "឵",
    "⠀",
    "ㅤ",
    "ﾠ",
    "𝅙",
    "\U0001d173",
    "\U0001d174",
    "\U0001d175",
    "\U0001d176",
    "\U0001d177",
    "\U0001d178",
    "\U0001d179",
    "\U0001d17a",
)


def char_to_surrogates(char: str) -> tuple[int, int]:
    """
    A surrogate pair is two chars that represent another character.
    Since UTF-16 only stores characters from 0 to 0xFFFF,
    chars past 0xFFFF need to be split into two codepoints below 0xFFFF.

    This is useful even in plaintext Unicode notation, because ``\\u1D105`` is not a single
    character, it's two (``ᴐ5``, not ``𝄅``).

    :param char:
        A single character.
    :returns:
        A surrogate pair, in codepoints of the surrogates.
    """
    codepoint = ord(char)
    codepoint -= 0x10000
    return 0xD800 + (codepoint >> 10), 0xDC00 + (codepoint & 0x3FF)


def surrogates_to_char(surrogates: tuple[int, int]) -> str:
    """
    Given a tuple of surrogate chars, return the single codepoint they combine to.

    :param surrogates:
        A tuple of two surrogate codepoints.
    :returns:
        A single character of the resulting surrogates.
    """
    return chr((surrogates[0] - 0xD800) * 0x400 + (surrogates[1] - 0xDC00) + 0x10000)


def is_char_invisible(char: str) -> bool:
    """
    Return if `char` would be invisible (might not have glyph info).

    A character is "invisible" if it:

    * Is in these categories: ``Cf, Cc, Zl, Zs, Zp``
    * | Is equal to these codepoints:
      | `2800, 034F, 115F, 1160, 17B4, 17B5, 3164, FFA0, 1D159, 1D174, 1D176, 1D177, 1D178, 1D17A`
    * Is private use

    You can visit `<https://invisible-characters.com/>`_ if you would like to see the list.

    .. warning::
        "Invisibility" is not a valid Unicode standard property.
        Do use utilize it outside this application for standardization purposes.

    :param char:
        A single character.
    :returns:
        If `char` is a spacing character.
    """
    return (
        unicodedata.category(char) in ("Cf", "Cc", "Zl", "Zs", "Zp")
        or char in INVISIBLE_CHARS
        or is_char_private_use(char)
    )


def is_char_private_use(char: str) -> bool:
    """
    Return if `char` is in a Private Use Area (PUA).

    A PUA is one of these codepoint ranges:

    * U+E000 to U+F8FF
    * U+F0000 to U+FFFFD
    * U+100000 to U+10FFFD

    :param char:
        A single character.
    :returns:
        If `char` is in a Private Use Area.
    """
    codepoint = ord(char)
    return (0xE000 <= codepoint <= 0xF8FF) or (0xF0000 <= codepoint <= 0xFFFFD) or (0x100000 <= codepoint <= 0x10FFFD)


def str_to_tags(string: str) -> str:
    """
    Given `string`, which should have only ASCII characters, turn it into that
    same string but every character is a Tag of itself, instead.

    See https://www.compart.com/en/unicode/block/U+E0000.

    :param string: Any string, should have ASCII characters.
    :returns: `string` but with ASCII characters replaced with their Tag equivalents.
    """
    return "".join(chr(ord(ch) + 0xE0000) for ch in string if ord(ch) < 0x7F)



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Compacting is the process of taking every "bitmap" character texture provided
by a Minecraft font provider and storing it in as little space as possible.

Generally, this means:

1. Finding the largest effective dimensions out of all the characters,
2. Expanding the dimensions of all other character textures to fit those new dimensions,
3. Creating a new texture that will fit all characters in those dimensions,
4. Sequencing them one-after-another while ignoring characters with a blank texture,
5. Creating a new charlist that corresponds with the compacted texture.

This contains functions for compacting any provider(s) into individual textures
(one for each provider), and for compacting a list of providers into a single, cohesive texture.
"""
import math

import PIL.Image
import mcfonts.utils.bitmap


def compact_providers(
    providers: list[mcfonts.providers.AnyVanillaProvider],
    chars_in_row: int = 0,
    cell_size: tuple[int, int] = (0, 0),
    square_cells: bool = True,
    include_empty_glyphs: bool = True,
) -> tuple[list[str], PIL.Image.Image, tuple[int, int]]:
    """
    Compact the textures and character arrays in `providers`.
    All padding (null, space) will be removed, and blank space will be removed.
    Providers that are not "bitmap" type are ignored.

    This will result in a sheet file with only textured characters.
    This function is different from :func:`compact_providers`,
    in that it will modify the maximum character based on all the providers.

    It returns a single tuple of the new charlist list,
    and the resulting compacted :class:`PIL.Image.Image`.

    :param providers:
        A list of :class:`mcfonts.providers.AnyProvider` instances,
        although only :class:`mcfonts.providers.BitmapProvider` instances will be considered.
    :param chars_in_row: How many characters to fit inside each row of the resulting sheet.
        Only positive values will set fixed rows.
        If this is 0, this will be set to the largest length of the first string in the
        charlist of all providers.
        If this is negative, this will be set so that the resulting sheet is square.
        By default, this is 0 (auto first string).
    :param cell_size:
        What size to make each glyph cell.
        If this is (0, 0), this will be set to the largest dimensions of every glyph in `glyphs`.
        If this is any other tuple of numbers,
    :param square_cells:
        If True, each glyph's width will equal its height.
        This is based on whichever number is largest.
        If False, each glyph's width will be unrelated to its height.
    :param include_empty_glyphs:
        If True, None elements in `glyphs` will be skipped over but their presense will be shown
        in the resulting file (empty space).
        If False, they will be filtered out entirely before being used.
    :returns:
        A list of the new characters, and a tuple containing a modified `resource`
        with the least amount of padding between glyphs,
        plus a tuple of the size of each glyph cell.
    """
    extracted = {}
    for provider in providers:
        if isinstance(provider, mcfonts.providers.BitmapProvider):
            extracted.update(provider.glyphs)

    if chars_in_row == 0:
        chars_in_row = max(
            len(provider.contents["chars"][0] or "")
            for provider in providers
            if isinstance(provider, mcfonts.providers.BitmapProvider)
        )
    elif chars_in_row < 0:
        chars_in_row = math.ceil(math.sqrt(len(extracted)))

    compacted = compact_glyphs(list(extracted.values()), chars_in_row, cell_size, square_cells, include_empty_glyphs)
    return (
        list(mcfonts.utils.fit_chars_into_charlist(list(extracted.keys()), chars_in_row)),
        compacted[0],
        compacted[1],
    )


def compact_glyphs(
    glyphs: list[PIL.Image.Image | None],
    chars_in_row: int = 0,
    cell_size: tuple[int, int] = (0, 0),
    square_cells: bool = True,
    include_empty_glyphs: bool = True,
) -> tuple[PIL.Image.Image, tuple[int, int]]:
    """
    Given a list of glyphs,
    return an image wherein all glyphs are compacted into a single sheet,
    with the number of glyphs in one row matches ``chars_in_row`` (excluding exceptions).

    :param glyphs: A list of :class:`PIL.Image.Image`.
    :param chars_in_row: How many characters to fit inside each row of the resulting sheet.
        Only positive values will set fixed rows.
        If this is negative or 0, this will be set so that the resulting sheet is square.
        By default, this is 0 (square).
    :param cell_size:
        What size to make each glyph cell.
        If this is (0, 0), this will be set to the largest dimensions of every glyph in `glyphs`.
        If this is any other tuple of numbers,
    :param square_cells:
        If True, each glyph's width will equal its height.
        This is based on whichever number is largest.
        If False, each glyph's width will be unrelated to its height.
    :param include_empty_glyphs:
        If True, None elements in `glyphs` will be skipped over but their presense will be shown
        in the resulting file (empty space).
        If False, they will be filtered out entirely before being used.
    :returns:
        A tuple containing a modified `resource` with the least amount of padding between glyphs,
        plus a tuple of the size of each glyph cell.
    """
    if not include_empty_glyphs:
        glyphs = list(glyph for glyph in glyphs if glyph is not None)
    if chars_in_row <= 0:
        chars_in_row = math.ceil(math.sqrt(len(glyphs)))

    if cell_size == (0, 0):
        max_cell = mcfonts.utils.bitmap.get_largest_effective_glyph_dimensions(glyphs)
    else:
        max_cell = cell_size
    if square_cells:
        maximum = max(max_cell)
        max_cell = (maximum, maximum)

    destination = PIL.Image.new(
        "RGBA",
        (max_cell[0] * chars_in_row, (max_cell[1] * math.ceil(len(glyphs) / chars_in_row)) + 1),
        (0, 0, 0, 0),
    )
    for index, glyph in enumerate(glyphs):
        if glyph:
            destination.paste(
                glyph,
                (
                    (index % chars_in_row) * max_cell[0],
                    (index // chars_in_row) * max_cell[1],
                ),
            )
    return destination, max_cell



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Utilities for constructing and translating rangestrings.
See :doc:`rangestrings` for details on the syntax.
"""
import mcfonts.constants
import mcfonts.exceptions


def part_to_character(part_notation: str) -> str:
    """
    Given a part of a rangestring, return the character it translates to.

    For details on what this notation should be, see :doc:`rangestrings`.
    This notation should match this regular expression
    :const:`mcfonts.constants.RANGESTRING_PART`.

    >>> part_to_character("U+2600")
    '☀'
    >>> part_to_character("0x2541")
    '╁'

    :param part_notation: A rangestring part.
    :returns: A single character.
    :raises mcfonts.exceptions.RangestringError: If the part is not valid.
    """
    if parts := mcfonts.constants.RANGESTRING_PART.fullmatch(part_notation):
        if (single_char := parts.group(5)) is not None:
            return single_char
        return chr(int(parts.group(4), 16))
    raise mcfonts.exceptions.RangestringError("Invalid part in rangestring")


def range_to_update(range_notation: str, current_coverage: set[str] | None = None) -> None:
    """
    Given a range of a rangestring, update `current_coverage` with the characters it translates to.

    For details on what this notation should be, see :doc:`rangestrings`.
    This notation should match this regular expression
    :const:`mcfonts.constants.RANGESTRING_RANGE`.

    :param range_notation:
        A rangestring range.
        This function does not accept individual parts.
    :param current_coverage:
        A set of the already-included characters.
        This is required for applying exclusion.
    :returns: Nothing, ``current_coverage`` is updated in place.
    :raises mcfonts.exceptions.RangestringError:
        If the range is not valid,
        or if a single part is supplied instead.
    """
    if current_coverage is None:
        current_coverage = set()
    split = mcfonts.constants.RANGESTRING_RANGE_DELIMITER.split(range_notation, 1)
    len_split = len(split)
    if len_split == 3:
        complete = set(
            chr(i)
            for i in range(
                ord(part_to_character(split[0])),
                # + 1 to be inclusive of last codepoint
                ord(part_to_character(split[2])) + 1,
            )
        )
        if split[0].startswith("!") or split[2].startswith("!"):
            # This is an exclusion, remove it
            current_coverage.difference(complete)
        else:
            current_coverage.update(complete)
    elif len_split == 2:
        raise mcfonts.exceptions.RangestringError("Implied infinite expansion is not allowed in rangestrings")
    # Didn't get two parts, not valid.
    else:
        raise mcfonts.exceptions.RangestringError("Invalid range in rangestring")


def rangestring_to_characters(rangestring: str) -> set[str]:
    r"""
    Return a set of characters that corresponds to `rangestring`.

    See :doc:`rangestrings` for more details.

    >>> rangestring_to_characters("U+2600..U+26ff,U+1ab0..0x1abf, 0x156, A")
    [['Ŗ', 'A'], range(9728, 9984), range(6832, 6848)]

    :param rangestring: A rangestring.
    :returns: A set of characters.
    """
    current_coverage: set[str] = set()
    for component in mcfonts.constants.RANGESTRING_COMPONENT_DELIMITER.split(rangestring):
        if mcfonts.constants.RANGESTRING_RANGE.fullmatch(component):
            # It's a range
            range_to_update(component, current_coverage)
        elif mcfonts.constants.RANGESTRING_PART.fullmatch(component):
            # It's a part
            if component.startswith("!"):
                # This is an exclusion, remove it
                current_coverage.difference_update(part_to_character(component))
            else:
                current_coverage.update(part_to_character(component))
        elif mcfonts.constants.RANGESTRING_RANGE_DELIMITER.search(component):
            # We didn't match the full range notation but there's a range delimiter in here
            # assume it's A- or -A notation; raise error.
            raise mcfonts.exceptions.RangestringError("Implied infinite expansion is not allowed in rangestrings")
        else:
            # It's neither, ???
            raise mcfonts.exceptions.RangestringError("Invalid rangestring notation")
    return current_coverage



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
CLI front-end for mcfonts.
"""
import argparse
import sys

import PIL.Image
import mcfonts.importing
import mcfonts.utils.bitmap

if __name__ != "__main__":
    raise NotImplementedError

INPUT_HELP = (
    "Path to folder, JSON, or ZIP file of font. "
    "If folder, this folder must be a resource pack. "
    "If JSON, this file must be a font JSON file. "
    "If ZIP, this file must be a resource pack archive. "
    "If -, stdin is used."
)

parser = argparse.ArgumentParser(
    description="""
    mcfonts is a versatile, fast, and extensible package for working with Minecraft fonts.
    
    The CLI front-end does not expose every possible option.
    For more in-depth usage, import the mcfonts module in Python.
    """
)

parser.add_argument(
    "-v",
    "--version",
    action="version",
    version="mcfonts " + ".".join(str(x) for x in mcfonts.__version__),
)

subparsers = parser.add_subparsers(title="Functions")

export_parser = subparsers.add_parser("export", help="export a font file to an OpenType font")
export_parser.add_argument(
    "input",
    type=argparse.FileType(),
    default=(None if sys.stdin.isatty() else sys.stdin),
    help=INPUT_HELP,
)
export_parser.add_argument("outfile", type=argparse.FileType("wb"), default=sys.stdout)
export_parser.add_argument(
    "-c",
    "--color",
    default=False,
    action="store_true",
    help="To read colors in referenced resources",
)
export_parser.add_argument("-n", "--font-name", help="What the font name of the output should be")
export_parser.set_defaults(export=True)


info_parser = subparsers.add_parser("info", help="print information about a font file")
info_parser.add_argument(
    "input",
    type=argparse.FileType(),
    default=(None if sys.stdin.isatty() else sys.stdin),
    help=INPUT_HELP,
)

info_parser.add_argument(
    "-t",
    "--table-chars",
    action="store_true",
    default=False,
    help="to put characters into a formatter table or put in single list",
)
info_parser.add_argument(
    "-s",
    "--summary-only",
    action="store_true",
    default=False,
    help="to print only a small summary",
)
info_parser.set_defaults(info=True)

compact_parser = subparsers.add_parser("compact", help="fit the glyphs in the minimum amount of space needed")
compact_parser.add_argument(
    "input",
    type=argparse.FileType("rb"),
    default=(None if sys.stdin.isatty() else sys.stdin),
    help="Path to PNG file",
)
compact_parser.add_argument(
    "glyphsize",
    nargs=2,
    type=int,
    default=(8, 8),
    help="Two numbers of the size of 1 glyph in `input`, (width, height)",
)
compact_parser.add_argument(
    "-c",
    "--chars-in-row",
    type=int,
    default=16,
    help="The desired number of glyphs to put in one row, set to 0 for square.",
)
compact_parser.add_argument("outfile", type=argparse.FileType("wb"), default=sys.stdout)
compact_parser.set_defaults(compact=True)


compare_parser = subparsers.add_parser("compare", help="compare two fonts")
compare_parser.add_argument(
    "input1",
    type=argparse.FileType("rb"),
    default=(None if sys.stdin.isatty() else sys.stdin),
    help=INPUT_HELP,
)
compare_parser.add_argument(
    "input2",
    type=argparse.FileType("rb"),
    default=(None if sys.stdin.isatty() else sys.stdin),
    help=INPUT_HELP,
)
compare_parser.add_argument(
    "-s",
    "--swap",
    action="store_true",
    default=(None if sys.stdin.isatty() else sys.stdin),
    help="swap the order of fonts compared",
)
compare_parser.set_defaults(compare=True)

vargs = vars(parser.parse_args(args=None if sys.argv[1:] else ["--help"]))

if vargs.get("export", False):
    mcfonts.importing.from_java_font_file(vargs["input"].name, vargs["color"]).export(vargs["font_name"]).save(
        vargs["outfile"], False
    )
elif vargs.get("info", False):
    mcfonts.importing.from_java_ambiguous(vargs["input"].name).print_info(
        vargs.get("table_chars", False), vargs.get("summary_only", False)
    )
elif vargs.get("compact"):
    resource = PIL.Image.open(vargs["input"])
    cell_size: tuple[int, int] = tuple(vargs["glyphsize"])
    glyphs = list(mcfonts.utils.bitmap.get_glyphs_from_resource(resource, cell_size))
    compacted = mcfonts.compacting.compact_glyphs(
        glyphs,
        vargs["chars_in_row"](resource.width // cell_size[0], resource.height // cell_size[1]),
    )
    compacted[0].save(vargs["outfile"])
    print(compacted[1])
elif vargs.get("compare"):
    font1 = mcfonts.importing.from_java_ambiguous(vargs["input1"].name)
    font2 = mcfonts.importing.from_java_ambiguous(vargs["input2"].name)
    if vargs.get("swap", False):
        font2.compare(font1)
    else:
        font1.compare(font2)



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""Contains the Provider classes as well the basic Provider class that they all inherit from."""
import math
import typing
import unicodedata

import jsonschema
import lxml.etree
import PIL.Image

import mcfonts.constants
import mcfonts.exceptions
import mcfonts.exporting
import mcfonts.unicode
import mcfonts.utils
import mcfonts.utils.bitmap
import mcfonts.utils.rangestring


class Provider:
    """The base provider class."""

    def __init__(self, provider_type: str, contents: dict, chars_covered: set[str] | None = None):
        if chars_covered is None:
            chars_covered = set()
        self.provider_type: str = provider_type
        self.chars_covered: set[str] = chars_covered
        self.contents: dict = contents

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(" f'"{self.provider_type}", {self.chars_covered}, {self.contents})'

    def __str__(self) -> str:
        return f"{self.__class__.__name__}({len(self.chars_covered)} chars)"

    def validate(self) -> None:
        """
        Specific tests for validating a provider.

        Values are checked to ensure that they're correct and in an acceptable range.
        Textures are checked to ensure all characters have a defined glyph,
        and that the dimensions are correct.

        Warnings are printed through :func:`logger.warning`.

        :returns: Nothing; problems are either raised or warned.
        """

    def print_info(self) -> None:
        """Print information about this provider."""
        print(f"Provider type: {self.provider_type or mcfonts.constants.UNKNOWN_FIELD}")
        print(f"Characters covered: {len(self.chars_covered)}")


class OptionsProvider(Provider):
    """A mcfonts-specific provider that only is used when exporting."""

    def __init__(self, provider: dict):
        self.contents = provider.copy()
        self.contents["fields"] = {}

        # Expand the rangestrings, copy over fields
        for field in mcfonts.constants.OPTION_FIELDS:
            if field not in self.contents["fields"]:
                self.contents["fields"][field] = {}
            for key, value in provider.get("fields", {}).get(field, {}).items():
                for newchar in mcfonts.utils.rangestring.rangestring_to_characters(key):
                    self.contents["fields"][field][newchar] = value
        chars_covered: set[str] = set()
        for field in mcfonts.constants.OPTION_FIELDS:
            chars_covered.update(self.contents["fields"].get(field, {}).keys())
        # Explicitly put mcfonts:options
        super().__init__("mcfonts:options", self.contents, chars_covered)
        self.validate()

    def validate(self) -> None:
        """
        Validate a bitmap provider,
        ensuring its structure is correct and that all values are acceptable.

        This function does validate textures.

        :returns: Nothing; problems are either raised or warned.
        """
        jsonschema.validate(self.contents, mcfonts.constants.SCHEMA_PROVIDER_OPTIONS)
        for char in self.chars_covered:
            if self.contents.get("fullwidth", {}).get(char) and self.contents.get("spacing", {}).get(char):
                mcfonts.logger.warning(
                    format_provider_message(
                        self, f"has conflicting options, fullwidth and spacing are present for character {char}."
                    )
                )
            if self.contents.get("fullwidth", {}).get(char) and self.contents.get("width", {}).get(char):
                mcfonts.logger.warning(
                    format_provider_message(
                        self, f"has conflicting options, fullwidth and width are present for character {char}."
                    )
                )
            if self.contents.get("width", {}).get(char) and self.contents.get("spacing", {}).get(char):
                mcfonts.logger.warning(
                    format_provider_message(
                        self, f"has conflicting options, width and spacing are present for character {char}."
                    )
                )

    def get_options_for_char(self, char: str) -> dict[str, int | bool] | None:
        """
        Given `char`, a character, return a dictionary of all the options for that character.

        For example, if :attr:`self.contents` is::
            {"width": {"a": 4}, "spacing": {"a": 0}}
        Then `self.get_options_for_char("a")` should return::
            {"width": 4, "spacing": 0}

        :param char: A single character.
        :returns: A flat dictionary containing a value for each field that `char` has declared.
        """
        result = {}
        for field in mcfonts.constants.OPTION_FIELDS:
            if (char_options := self.contents.get(field, {}).get(char)) is not None:
                result[field] = char_options
        if result:
            return result
        return None


class BitmapProvider(Provider):
    """
    A bitmap provider.
    """

    def export(
        self,
        font_xml,
        already_allocated_chars: set[str],
        select_chars: set[str] | None = None,
        select_policy: typing.Literal["include", "exclude"] = "include",
        options: OptionsProvider | None = None,
        options_handling: typing.Literal["include", "ignore", "include_only", "ignore_only"] = "include",
    ) -> set[str]:
        """
        Export a bitmap provider to a font XML.

        Options are honored, regardless of whatever the exporting function above says.
        To disable this, set ``options = {}`` before passing it in.

        The font XML is modified in-place.

        :param font_xml: The font XML.
        :param already_allocated_chars: A set of the allocated chars in `font_xml`.
        :param select_chars:
            A set of characters that will be either included or excluded from the exported font.
            To use a rangestring with this, use :func:`mcfonts.utils.rangestring.rangestring_to_characters`.

            For how to control this selection's effect, see `select_policy`.
            If this is an empty set, it will be ignored, regardless of what `select_policy` is.
        :param select_policy:
            How to handle `select_chars`.

            * "include": Only characters that are in this set will be exported *(default)*.
            * "exclude": Characters that are in this set will not be exported.
        :param options:
            An optional :class:`OptionsProvider` instance.
            If supplied, exported characters will obey it.
        :param options_handling:
            How to handle :doc:`options` and `options`:

            * If "include", include characters that have options declared.
            * If "ignore", disregard the options for a char, even if declared (but still include the char)
            * If "include_only", only include characters that have options declared.
            * If "ignore_only", only include characters that don't have any options declared.
        :returns: A new set of the characters added only in this provider.
        """
        # Ensure there's work to be done (there's chars to export) before even trying
        # also ensure that there's a resource to grab glyphs from
        if (
            len(self.contents["chars"]) < 1
            or not mcfonts.utils.bitmap.assert_charlist_is_equal(self.contents["chars"])
            or not self.resource
        ):
            return set()
        allocated_chars = []
        if select_chars is None:
            select_chars = set()
        for charline in self.contents["chars"]:
            if mcfonts.utils.is_charline_empty(charline):
                continue
            for character in charline:
                if character in mcfonts.constants.PADDING_CHARS or (
                    (select_policy == "exclude" and character in select_chars)
                    or (select_policy == "include" and character not in select_chars)
                    and select_chars
                ):
                    continue
                char_options = None
                if options is not None and (char_options := options.get_options_for_char(character)):
                    # Options exist
                    if options_handling == "ignore":
                        # Disregard options anyway, leave char itself alone
                        char_options = None
                    elif options_handling == "ignore_only":
                        # Disregard the char entirely
                        continue
                    # If it isn't any of the above, it must be `include*`, so leave char_options as-is.
                elif options_handling == "include_only":
                    # Only using chars with options; disregard this char
                    continue
                if glyph := self.glyphs.get(character):
                    if character not in already_allocated_chars:
                        # Only allocate the character if it doesn't exist
                        mcfonts.exporting.allocate_char(font_xml, character)
                        allocated_chars.append(character)
                    if program := mcfonts.exporting.glyph_to_program(
                        glyph,
                        (self.contents["ascent"] - self.contents["height"])
                        * (self.glyph_cell[1] // self.contents["height"]),
                        # Em size (1000) divided by 8 (standard width) = 125,
                        # divide by scale to get how big each "pixel" will translate to
                        125 / (self.glyph_cell[1] / self.contents["height"]),
                        char_options,
                    ):
                        mcfonts.exporting.set_program_to_char(font_xml, program, character)
                    else:
                        # Nothing here, it was a space.
                        mcfonts.logger.debug(
                            f"Glyph for {mcfonts.utils.pretty_print_char(character)} is "
                            f"empty, replacing with a 1-wide space."
                        )
                        mcfonts.exporting.set_space_to_char(font_xml, character, 1)
        return set(allocated_chars)

    def __init__(self, provider: dict, resource: PIL.Image.Image | None):
        self.contents = provider
        # Default height fallback
        self.contents["height"] = self.contents.get("height", 8)
        self.resource = resource
        if self.resource:
            if len(self.contents["chars"]) > 0:
                self.glyph_cell = (
                    self.resource.width // len(self.contents["chars"][0] or ""),
                    self.resource.height // len(self.contents["chars"]),
                )
            else:
                self.glyph_cell = (self.resource.width, self.resource.height)
        else:
            self.glyph_cell = (0, 0)
        self.validate()

        # Ensure resource exists before calculating glyph cell box
        self.glyphs: dict[str, PIL.Image.Image] = dict(self.yield_glyphs())

        chars_covered = []
        for charline in self.contents["chars"]:
            for char in charline:
                if char not in mcfonts.constants.PADDING_CHARS:
                    chars_covered.append(char)
        super().__init__("bitmap", self.contents, set(chars_covered))

    def validate(self) -> None:
        """
        Validate a bitmap provider,
        ensuring its structure is correct and that all values are acceptable.

        This function does validate textures.

        :returns: Nothing; problems are either raised or warned.
        """
        jsonschema.validate(self.contents, mcfonts.constants.SCHEMA_PROVIDER_BITMAP)
        mcfonts.utils.bitmap.validate_charlist(self.contents["chars"], self.resource, self.glyph_cell)
        if self.resource:
            if self.glyph_cell[1] % self.contents["height"] != 0:
                mcfonts.logger.warning(
                    format_provider_message(self, "has an invalid height, it must be a multiple of the glyph's height.")
                )
            if self.glyph_cell[0] > 256 or self.glyph_cell[1] > 256:
                # The limit for a character is 256, anything beyond that is an error in
                # atlas-stitching
                # Will still try to make it work but makes no guarantee
                mcfonts.logger.warning(
                    format_provider_message(
                        self,
                        "has excessively large glyph size, maximum character cell size is 256, "
                        f"but got {max(self.glyph_cell[0], self.glyph_cell[1])}.",
                    )
                )

    def get_glyphs_in_rangestring(self, rangestring: str) -> typing.Iterator[tuple[str, PIL.Image.Image | None]] | None:
        """
        Given `rangestring`,
        return a dictionary of the requested characters to their glyphs.

        A rangestring is a comma-separated string of characters or range notations.
        Essentially, accepts any of these:

        * 16FF => one codepoint, 5887
        * 2000-22ff => range of codepoints from 8192 to 8959
        * U+5460..5800 => range of codepoints from 21600 to 22528
        * U+6000..6001 => range of codepoints from 24576 to 24577
        * A-Z => range of codepoints from 65 to 90
        * ☃ => one codepoint, 9731

        Ranges can be split by commas.

        :param rangestring:
            A string representing the requested range of chars.
            See :func:`mcfonts.utils.ranges_from_rangestring()` for details.
        :returns:
            A yield of the requested glyphs that match `rangestring`,
            or None if there are no such matching glyphs.
        """
        if not self.resource:
            return {}
        wanted: set[str] = mcfonts.utils.rangestring.rangestring_to_characters(rangestring)
        for y_index, charline in enumerate(self.contents["chars"]):
            if mcfonts.utils.is_charline_empty(charline):
                # Line is only spaces or null padding, ignore it entirely to save time
                continue
            for x_index, char in enumerate(charline):
                # Is character not null nor space and
                # is there an actual glyph there (parameters permitting)
                if char not in mcfonts.constants.PADDING_CHARS and char in wanted:
                    glyph = self.resource.crop(
                        (
                            x_index * self.glyph_cell[0],
                            y_index * self.glyph_cell[1],
                            (x_index + 1) * self.glyph_cell[0],
                            (y_index + 1) * self.glyph_cell[1],
                        )
                    )
                    if not mcfonts.utils.is_image_empty(glyph):
                        yield char, glyph
                    else:
                        yield char, None
        return {}

    def print_info(self, table_chars: bool = True) -> None:
        """
        Print information about this provider.

        :param table_chars:
            Whether to print a 'chars' list as a square table, or as a simple string.
        """
        super().print_info()
        print(f"File: {self.contents.get('file', mcfonts.constants.UNKNOWN_FIELD)}")
        print(f"Height: {self.contents.get('height', mcfonts.constants.UNKNOWN_FIELD)}")
        print(f"Ascent: {self.contents.get('ascent')}")
        if table_chars:
            print(f"Chars: ({len(self.contents['chars'][:1]):,}x{len(self.contents['chars']):,})")
            for charline in self.contents["chars"]:
                print(f"\t{' '.join(charline)}")
        else:
            print(f"Chars: {' '.join(''.join(self.contents['chars']))}")
        print(f"Count: {sum(len(x) for x in self.contents['chars']):,}")

    def yield_glyphs(
        self,
        ignore_empty_textures: bool = False,
    ) -> typing.Iterator[tuple[str, PIL.Image.Image | None]]:
        """
        Extract every textured glyph from this provider.
        Optionally, include the height and ascent.

        :param ignore_empty_textures: Whether to not include characters that have no pixel data.
        :returns: A yield of {codepoint: glyph | None}.
        """
        if not self.resource:
            return {}
        for y_index, charline in enumerate(self.contents["chars"]):
            if mcfonts.utils.is_charline_empty(charline):
                # Line is only spaces or null padding, ignore it entirely to save time
                continue
            for x_index, char in enumerate(charline):
                # Is character not null nor space and
                # is there an actual glyph there (parameters permitting)
                if char not in mcfonts.constants.PADDING_CHARS:
                    glyph = self.resource.crop(
                        (
                            x_index * self.glyph_cell[0],
                            y_index * self.glyph_cell[1],
                            (x_index + 1) * self.glyph_cell[0],
                            (y_index + 1) * self.glyph_cell[1],
                        )
                    )
                    empty = mcfonts.utils.is_image_empty(glyph)
                    if empty or not ignore_empty_textures:
                        yield char, glyph
                    elif empty and ignore_empty_textures:
                        yield char, None
        return {}

    def get_glyph_for_char(self, char: str) -> PIL.Image.Image | None:
        """
        Extract a desired character's sheet from this provider.

        Glyph textures that are mapped to spaces or null padding return None.
        If there is more than one sheet mapped to a character (duplicate characters),
        the first occurance will be returned.

        .. warning::
            This method is not efficient. Do not use it repeatedly,
            use :func:`get_glyphs_from_providers` and get each glyph from there.

        :param char: A single character.
        :returns: A :class:`PIL.Image.Image`, or None if no glyph exists.
        """
        # Ignore non-bitmap provider or if codepoint is null or space
        # (glyphs can't be assigned for padding characters)
        if char in mcfonts.constants.PADDING_CHARS or self.resource is None:
            return None
        for y_index, charline in enumerate(self.contents["chars"]):
            if mcfonts.utils.is_charline_empty(charline):
                # Line is only spaces or null padding, ignore it entirely to save time
                continue
            for x_index, character in enumerate(charline):
                if char == character:
                    return self.resource.crop(
                        (
                            x_index * self.glyph_cell[0],
                            y_index * self.glyph_cell[1],
                            (x_index + 1) * self.glyph_cell[0],
                            (y_index + 1) * self.glyph_cell[1],
                        )
                    )
        return None

    def reload_to_monochrome(self) -> None:
        """
        Replace the resource used in this provider with a grayscale version.
        If the resource is already grayscale, this will have no effect.

        This modifies the resource of this provider in place, and **cannot be undone**.
        """
        if self.resource:
            if self.resource.mode == "LA":
                return
            self.resource = self.resource.convert("LA")


class SpaceProvider(Provider):
    """A space provider."""

    def __init__(self, provider: dict):
        self.contents = provider.copy()
        self.contents["advances"] = {}
        # Expand the rangestrings, copy over fields
        for key, value in provider.get("advances", {}).items():
            for newchar in mcfonts.utils.rangestring.rangestring_to_characters(key):
                self.contents["advances"][newchar] = value
        self.validate()
        super().__init__("space", provider, set(self.contents["advances"].keys()))

    def export(
        self,
        font_xml: lxml.etree._Element,
        already_allocated_chars: set[str],
        select_chars: set[str] | None = None,
        select_policy: typing.Literal["include", "exclude"] | None = None,
    ) -> set[str]:
        """
        Export a space provider to a font XML.
        The font XML is modified in-place.

        :param font_xml: The font XML.
        :param already_allocated_chars: A set of the allocated chars in `font_xml`.
        :param select_chars:
            A set of characters that will be either included or excluded from the exported font.
            To use a rangestring with this, use :func:`mcfonts.utils.rangestring.rangestring_to_characters`.

            For how to control this selection's effect, see `select_policy`.
            If this is an empty set, it will be ignored, regardless of what `select_policy` is.
        :param select_policy:
            How to handle `select_chars`.

            * "include": Only characters that are in this set will be exported.
            * "exclude": Characters that are in this set will not be exported.
        :returns: A new set of the characters added only in this provider.
        """
        allocated_chars = []
        if select_chars is None:
            select_chars = set()
        for character, width in self.contents["advances"].items():
            if (
                (select_policy == "exclude" and character in select_chars)
                or (select_policy == "include" and character not in select_chars)
                and select_chars
            ):
                continue
            if math.isinf(width) or math.isnan(width):
                # Some values may be infinite, limit them to 20
                width = 20
            if character not in already_allocated_chars:
                # Only allocate the character if it doesn't exist
                mcfonts.exporting.allocate_char(font_xml, character)
                allocated_chars.append(character)
            mcfonts.exporting.set_space_to_char(font_xml, character, abs(int(width)))
        return set(allocated_chars)

    @staticmethod
    def to_glyph_sizes(advances: dict[str, int], default_width: tuple[int, int] = (0, 14)) -> bytearray:
        """
        Create a glyph_sizes.bin bytearray from a template of chars and their starting and
        ending positions.

        :param advances: A dictionary of `{character: width}`.
        :param default_width:
            The width to fall back to if `advances` does not define one for a character.
        :returns: Bytearray of glyph_sizes.bin.
        """
        glyphsizes = bytearray((default_width[0] * 16 + default_width[1]).to_bytes(1, "big") * 65536)
        for character, width in advances.items():
            if (codepoint := ord(character)) > 0xFFFF:
                # Cannot pack characters higher than the BMP
                mcfonts.logger.warning(
                    f"Cannot include character {character} in glyph_sizes; " "codepoint is above U+FFFF."
                )
                continue
            # Ensure the high and low bits are in correct 0-F range
            if 0 > width < 15:
                raise ValueError("Width must be within 0 to 15")
            glyphsizes[codepoint] = width
        return glyphsizes

    def validate(self) -> None:
        """
        Validate a "space" provider,
        ensuring its structure is correct and that all values are acceptable.

        :returns: Nothing; problems are either raised or warned.
        """
        jsonschema.validate(self.contents, mcfonts.constants.SCHEMA_PROVIDER_SPACE)

    def print_info(self) -> None:
        """
        Print information about this provider.
        """
        super().print_info()

        if len(self.contents["advances"]) < 1:
            # There's no advances
            print("No advances.")
        else:
            advances = self.contents["advances"]
            print(f"Advances: ({len(advances)})")
            for spacechar, width in advances.items():
                print(f"\tCharacter {mcfonts.utils.pretty_print_char(spacechar)}: {width}")


class TTFProvider(Provider):
    """
    A TTF provider.
    """

    def __init__(self, provider: dict, resource: bytes | None = None):

        self.contents = provider
        if "skip" in self.contents:
            chars = []
            if isinstance(skip := self.contents.get("skip"), list):
                for line in skip:
                    chars.append(line)
            # Flatten into a string
            self.contents["skip"] = "".join(chars)
        self.validate()
        self.resource = resource
        super().__init__("ttf", provider)

    def validate(self) -> None:
        """
        Validate a "ttf" provider,
        ensuring its structure is correct and that all values are acceptable.

        .. warning:: This function does not validate textures.

        :returns: Nothing; problems are either raised or warned.
        :raises:
            :exc:`mcfonts.exceptions.MissingResourceError` if no resource was given
        """
        jsonschema.validate(self.contents, mcfonts.constants.SCHEMA_PROVIDER_TTF)


class LegacyUnicodeProvider(Provider):
    """A legacy unicode provider."""

    def __init__(self, provider: dict, resources: dict[str, PIL.Image.Image | bytes]):
        self.contents = provider
        self.validate()
        self.resources = {}
        # Only take resources that are relevant
        split = self.contents["template"].split("%s", 1)
        for name in resources:
            if (name.startswith(split[0]) and name.endswith(split[1])) or name == self.contents["sizes"]:
                self.resources[name] = resources[name]
        self.codepages_covered: list[int] = []
        super().__init__(
            "legacy_unicode",
            self.contents,
            # U+0000 to U+FFFF.
            set(chr(i) for i in range(65535)),
        )

    def export(
        self,
        font_xml: lxml.etree._Element,
        already_allocated_chars: set[str],
        select_chars: set[str] | None = None,
        select_policy: typing.Literal["include", "exclude"] = "include",
        options: OptionsProvider | None = None,
        options_handling: typing.Literal["include", "ignore", "include_only", "ignore_only"] = "include",
    ) -> set[str]:
        """
        Export a legacy unicode provider to a font XML.
        The font XML is modified in-place.

        Despite Vanilla behavior, this function **does not** make assumptions about dimensions or
        scale.

        :param font_xml: The font XML.
        :param already_allocated_chars: A set of the allocated chars in `font_xml`.
        :param select_chars:
            A set of characters that will be either included or excluded from the exported font.
            To use a rangestring with this, use :func:`mcfonts.utils.rangestring.rangestring_to_characters`.

            For how to control this selection's effect, see `select_policy`.
            If this is an empty set, it will be ignored, regardless of what `select_policy` is.
        :param select_policy:
            How to handle `select_chars`.

            * "include": Only characters that are in this set will be exported *(default)*.
            * "exclude": Characters that are in this set will not be exported.
        :param options:
            An optional :class:`OptionsProvider` instance.
            If supplied, exported characters will obey it.
        :param options_handling:
            How to handle :doc:`options` and `options`:

            * If "include", include characters that have options declared.
            * If "ignore", disregard the options for a char, even if declared (but still include the char)
            * If "include_only", only include characters that have options declared.
            * If "ignore_only", only include characters that don't have any options declared.
        :returns: A new set of the characters added only in this provider.
        """
        allocated_chars = []
        if select_chars is None:
            select_chars = set()
        for unicode_codepage in range(255):
            if unicode_codepage in range(216, 223):
                # Codepage is in surrogate range, there must be nothing here
                continue
            template_parts: list[str] = self.contents["template"].split("%s")
            mcfonts.logger.info(f"Working on Unicode page {hex(unicode_codepage).upper()}.")
            resource: PIL.Image.Image = self.resources.get(
                f"{template_parts[0]}{unicode_codepage:02x}{template_parts[1]}"
            )
            if resource is None:
                mcfonts.logger.warning(f"Provider has missing unicode page {unicode_codepage:02x}, ignoring.")
                continue
            # Calculate each character's bounding box
            resolution: tuple[int, int] = (
                resource.width // 16,
                resource.height // 16,
            )
            scale = resource.height / resolution[1]
            for y_index in range(16):
                for x_index in range(16):
                    character = chr(unicode_codepage * 256 + y_index * 16 + x_index)
                    if (
                        (select_policy == "exclude" and character in select_chars)
                        or (select_policy == "include" and character not in select_chars)
                        and select_chars
                    ):
                        continue
                    char_options = None
                    if options is not None and (char_options := options.get_options_for_char(character)):
                        # Options exist
                        if options_handling == "ignore":
                            # Disregard options anyway, leave char itself alone
                            char_options = None
                        elif options_handling == "ignore_only":
                            # Disregard the char entirely
                            continue
                        # If it isn't any of the above, it must be `include*`, so leave char_options as-is.
                    elif options_handling == "include_only":
                        # Only using chars with options; disregard this char
                        continue
                    if program := mcfonts.exporting.glyph_to_program(
                        resource.crop(
                            (
                                x_index * resolution[0],
                                y_index * resolution[1],
                                (x_index + 1) * resolution[0],
                                (y_index + 1) * resolution[1],
                            )
                        ),
                        -2,
                        125 / scale,
                        char_options,
                    ):
                        if character not in already_allocated_chars:
                            # Only allocate the character if it doesn't exist
                            mcfonts.exporting.allocate_char(font_xml, character)
                            allocated_chars.append(character)
                        mcfonts.exporting.set_program_to_char(
                            font_xml,
                            program,
                            character,
                            False,
                        )
        return set(allocated_chars)

    def validate(self) -> None:
        """
        Validate a "legacy_unicode" provider,
        ensuring its structure is correct and that all values are acceptable.

        .. warning:: This function does not validate textures.

        :returns: Nothing; problems are either raised or warned.
        """
        jsonschema.validate(self.contents, mcfonts.constants.SCHEMA_PROVIDER_LEGACY_UNICODE)

    def print_info(self) -> None:
        """
        Print information about this provider.
        """
        print(f"Template: {self.contents['template'] or mcfonts.constants.UNKNOWN_FIELD}")
        print(f"Sizes: {self.contents['sizes'] or mcfonts.constants.UNKNOWN_FIELD}")

    @staticmethod
    def to_advances(glyphsizes: bytes, match_unicode_category: list[str] | None = None) -> dict[str, int]:
        """
        Translate a glyphsizes.bin file into an "advances" mapping, which goes inside a "space"
        provider.

        .. warning:: This function does not return a new provider.

        :param glyphsizes: The bytes of glyphsizes.bin.
        :param match_unicode_category: Only translate Unicode characters with these categories.
            By default, this is [Mc, Zp, Zs, Zl].
            This should cover most whitespace and marking characters.
        :returns: An "advances" dictionary.
        """
        if match_unicode_category is None:
            match_unicode_category = ["Mc", "Zp", "Zs", "Zl"]
        advances = {}
        for index, nibble in enumerate(glyphsizes):
            if unicodedata.category(char := chr(index)) in match_unicode_category:
                advances[char] = (nibble & 0xF) - (nibble >> 4 & 0xF)
        return advances


AnyProvider = typing.TypeVar("AnyProvider", bound=Provider)
"""
A TypeVar for any kind of Provider subclass.
This includes :class:`OptionsProvider`.
"""
AnyVanillaProvider = typing.TypeVar(
    "AnyVanillaProvider", BitmapProvider, SpaceProvider, LegacyUnicodeProvider, TTFProvider
)
"""
A TypeVar for any kind of Provider subclass that the Vanilla game accepts.
This does not include :class:`OptionsProvider`.
"""


def build_provider_list(
    providers: list[dict],
    resources: dict[str, PIL.Image.Image | bytes] | None = None,
    strict: bool = True,
) -> list[AnyProvider]:
    """
    Given a list of provider dictionaries, return a list of the appropriate :class:`Provider`
    classes.

    If an unknown provider is encountered, *it will not* be included.

    :param providers: A list of dictionaries that are individual providers.
    :param resources: A dictionary of {file name: :class:`PIL.Image.Image`}.
    :param strict:
        If a provider has bad data,
        an exception will be raised and no provider list will be returned if this is True.
        If this is False, None will be put in its place.
    :returns: A list of :class:`Provider` equivalents.
    :raises FileNotFoundError: If `strict` and a file is missing.
    :raises jsonschema.ValidationError: If `strict` and a provider structure is invalid
    """
    result: list[AnyProvider] = []
    if resources is None:
        resources = {}
    options = {}
    for provider in providers:
        try:
            if (provider_type := provider.get("type", "")).lower() == "bitmap":
                result.append(BitmapProvider(provider, resources.get(provider.get("file"))))
            elif provider_type == "space":
                result.append(SpaceProvider(provider))
            elif provider_type == "ttf":
                result.append(TTFProvider(provider, resources.get(provider.get("file"))))
            elif provider_type == "legacy_unicode":
                result.append(LegacyUnicodeProvider(provider, resources))
            elif provider_type in ("options", "mcfonts:options"):
                options = provider
        except (FileNotFoundError, jsonschema.ValidationError) as exception:
            if strict:
                raise exception
            continue
    try:
        if options:
            result.append(OptionsProvider(options))
    except jsonschema.ValidationError as exception:
        if strict:
            raise exception
    return result


def get_char_glyph_from_providers(
    providers: list[BitmapProvider],
    char: str,
) -> PIL.Image.Image | None:
    """
    Extract a desired character's glyph image from a list of
    :class:`mcfonts.providers.BitmapProvider`.

    Glyph textures that are mapped to spaces or null padding return None.
    If there is more than one sheet mapped to a character (duplicate characters),
    the first occurance will be returned.

    .. warning::
        This method is not efficient. Do not use it repeatedly,
        use :func:`get_glyphs_from_providers` and get each glyph from there.

    :param providers: The list of "bitmap" providers.
    :param char: A single character.
    :returns: A :class:`PIL.Image.Image`, or None if no glyph exists.
    """
    for provider in providers:
        if isinstance(provider, mcfonts.providers.BitmapProvider) and (glyph := provider.get_glyph_for_char(char)):
            return glyph
    return None


def format_provider_message(provider: AnyProvider | dict, message: str) -> str:
    """
    Format a provider warning message properly, following Provider <type>: <message>.

    Calls :func:`mcfonts.utils.pretty_print_provider()`.

    :param provider:
        The provider, either as a dictionary or instance of :data:`mcfonts.providers.AnyProvider`.
    :param message: The message to append at the end.
    :returns: The formatted warning message.
    """
    if isinstance(provider, dict):
        return f"Provider {pretty_print_provider_dictionary(provider)} {message}"
    return f"Provider {pretty_print_provider(provider)} {message}"


def pretty_print_provider(provider: AnyProvider) -> str:
    """
    Format a provider information message properly, following ``<type>`` plus:
     * ``bitmap``: ``<file> h <height> a <ascent>``
     * ``space``: nothing
     * ``legacy_unicode``: ``<template>``
     * ``ttf``: ``<file> s <shift0, 1>, sz <size>, o <oversample>, sk <skip>``

    :param provider:
        The provider as an instance of :data:`mcfonts.providers.AnyProvider`,
        or as a dictionary.
    :returns: The pretty provider information.
    """
    if isinstance(provider, BitmapProvider):
        return f"\"bitmap\": {provider.contents.get('file', 'no resource')}"
    if isinstance(provider, SpaceProvider):
        return '"space"'
    if isinstance(provider, LegacyUnicodeProvider):
        return f'"legacy_unicode": {provider.contents.get("template", "no template")}'
    if isinstance(provider, TTFProvider):
        return f'"ttf": {provider.contents.get("file", "no file")}'
    if isinstance(provider, OptionsProvider):
        return f'"options": {len(provider.chars_covered)} chars'
    return f'"{provider.provider_type}":'


def pretty_print_provider_dictionary(provider: dict) -> str:
    """
    Format a provider information message properly, following ``<type>`` plus:
     * ``bitmap``: ``<file> h <height> a <ascent>``
     * ``space``: nothing
     * ``legacy_unicode``: ``<template>``
     * ``ttf``: ``<file> s <shift0, 1>, sz <size>, o <oversample>, sk <skip>``

    :param provider:
        The provider as a dictionary,
        not an instance of :data:`mcfonts.providers.AnyProvider`.
    :returns: The pretty provider information.
    """
    if (provider_type := provider.get("type", "")).lower() == "bitmap":
        return f'"bitmap": {provider.get("file") or "no resource"}'
    if provider_type == "space":
        return '"space"'
    if provider_type == "legacy_unicode":
        return f'"legacy_unicode": {provider.get("template") or "no template"}'
    if provider_type == "ttf":
        return (
            f'"ttf": {provider.get("file", "no file")}, s '
            f'{provider.get("shift", "[?, ?]")}, sz '
            f'{provider.get("size", "?")}, o '
            f'{provider.get("oversample", "?")}, sk '
            f'{provider.get("skip", "none")}'
        )
    if provider_type == "options":
        return '"options"'
    return f'provider "{provider_type}": ?'



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Color codes and settings for mcfonts.

These are ANSI color codes.
Set :const:`USE_COLORS` to False to **disable** colored output.
Set :const:`USE_COLORS` to True to **enable** colored output.
"""

USE_COLORS: bool = True

_CSI = "\x1b["
BLACK_FORE = f"{_CSI}30m"
RED_FORE = f"{_CSI}31m"
GREEN_FORE = f"{_CSI}32m"
YELLOW_FORE = f"{_CSI}33m"
BLUE_FORE = f"{_CSI}34m"
MAGENTA_FORE = f"{_CSI}35m"
CYAN_FORE = f"{_CSI}36m"
WHITE_FORE = f"{_CSI}37m"
RESET_FORE = f"{_CSI}39m"

BLACK_BACK = f"{_CSI}40m"
RED_BACK = f"{_CSI}41m"
GREEN_BACK = f"{_CSI}42m"
YELLOW_BACK = f"{_CSI}43m"
BLUE_BACK = f"{_CSI}44m"
MAGENTA_BACK = f"{_CSI}45m"
CYAN_BACK = f"{_CSI}46m"
WHITE_BACK = f"{_CSI}47m"
RESET_BACK = f"{_CSI}49m"

RESET_ALL = f"{_CSI}0m"
BRIGHT = f"{_CSI}1m"
DIM = f"{_CSI}2m"



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Functions for exporting various character/provider formats into XML-representable data.

Contains functions for handling various providers and exporting them into font XMLs.
These providers are handled:

* space
* bitmap
* legacy_unicode
* mcfonts:options
"""
import datetime

import fontTools.pens.t2CharStringPen
import lxml.etree
import PIL.Image
import mcfonts.constants
import mcfonts.exceptions


def glyph_to_program(
    glyph: PIL.Image.Image, ascent: int, pixel_size: float, options: dict | None = None
) -> list[str] | None:
    """
    Create a Type 2 charstring program given a glyph sheet.
    If the glyph is empty or has no white pixels,
    None will be returned. Otherwise, a list of strings will be.

    These charstrings are **not** optimized or checked for overlaps.
    In FontForge, use :menuselection:`E&lement --> O&verlap --> &Remove Overlap` manually
    afterwards.

    :param glyph: A :class:`PIL.Image.Image` of the individual character, **not** the font sheet.
    :param ascent: The ascent of this glyph, equal to ``(ascent - height) * (glyph_height / height)``.
    :param pixel_size: How big each pixel should be.
    :param options:
        A specialized dictionary that should be derived
        from an instance of :class:`mcfonts.providers.OptionsProvider`.
        Should be in the form of {field*: value}. For example, an Options dictionary of
        ``{"width": {"a": 4}, "spacing": {"a": 2}}``
        should transform into ``{"width": 4, "spacing": 2}``.
        Use :meth:`mcfonts.providers.OptionsProvider.get_options_for_char` for this.
    :returns:
        A list of strings of the glyph's program,
        or None if there was no pixel data in ``glyph``.
    """
    # Excess is a number of how far away the glyph is from the left edge of the canvas
    if options is None:
        options = {}
    excess = get_glyph_nominals(glyph)
    # If excess is -1 (special case), there is nothing here but a space.
    if excess[1] == -1:
        return None
    # No components to resolve, so no glyphSet is needed; using {}
    pen = fontTools.pens.t2CharStringPen.T2CharStringPen(0, {})
    modifiers = [0, 0]
    width = excess[1]
    if options.get("fullwidth", False) is True:
        width = glyph.width
    # If shift is [0, 0], ignore it.
    if options.get("shift", [0, 0]) != [0, 0]:
        modifiers = options["shift"]
    if options.get("spacing") is not None:
        width = excess[1] + options["spacing"]
    if options.get("width") is not None:
        width = options["width"]
    if glyph.mode == "RGBA":
        data = glyph.getdata(3)
    elif glyph.mode == "LA":
        data = glyph.getdata(1)
    else:
        data = glyph.getdata(0)
    for index, pixel in enumerate(data):
        if pixel >= 180:
            x_position = index % glyph.width
            y_position = index // glyph.width
            pen.moveTo(
                (
                    (x_position + modifiers[0]) * pixel_size,
                    (glyph.height - y_position + ascent + modifiers[1]) * pixel_size,
                )
            )
            # Left X, top right
            pen.lineTo(
                (
                    (x_position + 1 + modifiers[0]) * pixel_size,
                    (glyph.height - y_position + ascent + modifiers[1]) * pixel_size,
                )
            )
            # Down Y, bottom left
            pen.lineTo(
                (
                    (x_position + 1 + modifiers[0]) * pixel_size,
                    (glyph.height - y_position - 1 + ascent + modifiers[1]) * pixel_size,
                )
            )
            # Right X, bottom right
            pen.lineTo(
                (
                    (x_position + modifiers[0]) * pixel_size,
                    (glyph.height - y_position - 1 + ascent + modifiers[1]) * pixel_size,
                )
            )
            # Done pixel
            pen.closePath()
    # Return the pen's resulting program
    charstring = pen.getCharString()
    # Set the width correctly
    charstring.program[0] = int(pixel_size * (width + 1))
    return charstring.program


def get_glyph_nominals(glyph: PIL.Image.Image) -> tuple[int, int]:
    """
    Return the number of pixels away from the left edge of the canvas the glyph is and the
    glyph's resulting calculated width.

    If return is 0, -1, there is no pixel data, the glyph is all spaces.

    :param glyph: A :class:`PIL.Image.Image` instance.
    :returns:
        Left padding and total width.
        Returns (0, -1) if no pixel data.
    """
    try:
        bbox = glyph.getbbox()
        return bbox[0], bbox[2]
    except (TypeError, IndexError):
        # Pure space, has no width
        return 0, -1


def allocate_char(font_xml: lxml.etree._Element, char: str) -> None:
    """
    Allocate a Unicode character in a font.
    Does not assign any character data or widths.

    :param font_xml: The font XML.
    :param char: A single character.
    :returns: Nothing, font is modified in-place.
    :raises GlyphLimitError: If there are more than 65,535 glyphs in the font; can't add more.
    """
    codepoint = ord(char)
    uni = f"u{codepoint:04X}"

    if (metrics_num := int((metrics := font_xml.find("hhea/numberOfHMetrics")).get("value"))) >= 65535:
        raise mcfonts.exceptions.GlyphLimitError
    metrics.set("value", str(metrics_num + 1))

    num_glyphs = font_xml.find("maxp/numGlyphs")
    num_glyphs.set("value", str(int(num_glyphs.get("value")) + 1))

    cmap = font_xml.find("cmap")

    if (cmap_12 := cmap.find("cmap_format_12")) is None:
        lxml.etree.SubElement(
            cmap,
            "cmap_format_12",
            {
                "platformID": "0",
                "platEncID": "4",
                "language": "0",
                "length": "4120",
                "nGroups": "323",
                "format": "12",
                "reserved": "0",
            },
        )
        cmap_12 = cmap.find("cmap_format_12")
    # Inside the BMP
    if codepoint <= 0xFFFF:
        # No format table exists yet, add it
        if (cmap_4 := cmap.find("cmap_format_4")) is None:
            lxml.etree.SubElement(
                cmap,
                "cmap_format_4",
                {
                    "platformID": "3",
                    "platEncID": "4",
                    "language": "0",
                },
            )
            cmap_4 = cmap.find("cmap_format_4")
        lxml.etree.SubElement(cmap_4, "map", {"code": f"0x{codepoint:X}", "name": uni, "language": "0"})
    lxml.etree.SubElement(cmap_12, "map", {"code": f"0x{codepoint:X}", "name": uni, "language": "0"})

    lxml.etree.SubElement(font_xml.find("GDEF/GlyphClassDef"), "ClassDef", {"glyph": uni, "class": "1"})

    lxml.etree.SubElement(font_xml.find("GlyphOrder"), "GlyphID", {"name": uni})


def set_program_to_char(font_xml: lxml.etree._Element, program: list[str], char: str, replace: bool = True) -> None:
    """
    Set a program to a character.
    This is how character data is added to the font.
    The character must be allocated already.
    That is not done in this function.

    If the character is not in the font, add it.
    If it is and ``replace`` is True, set it to the new character data.
    Otherwise, do nothing.

    :param font_xml: The font XML.
    :param program: A list of strings of the glyph's program.
    :param char: A single character.
    :param replace: If the character already has data, overwrite.
    :returns: Nothing, font is modified in-place.
    :raises GlyphLimitError: If there are more than 65,535 glyphs in the font; can't add more.
    """
    uni = f"u{ord(char):04X}"
    charstrings = font_xml.find("CFF/CFFFont/CharStrings")
    charstring_xml = font_xml.find(f"{charstrings}/CharString[@name='{uni}']")
    if charstring_xml is None:
        # Doesn't exist
        lxml.etree.SubElement(charstrings, "CharString", {"name": uni}).text = " ".join(str(x) for x in program)
        lxml.etree.SubElement(font_xml.find("hmtx"), "mtx", {"name": uni, "width": str(program[0]), "lsb": "0"})
    elif replace:
        # Exists already in the charstrings, replace it too
        charstring_xml.text = " ".join(str(x) for x in program)
        font_xml.find(f"hmtx/mtx[@name='{uni}']").set("width", str(program[0]))


def set_space_to_char(font_xml: lxml.etree._Element, char: str, width: int, replace: bool = True) -> None:
    """
    Add/set a whitespace character to the font, only defining its width.
    The charcter must be allocated already. That is not done in this function.

    If the glyph is not in the font, add it.
    If it is and ``replace`` is True, set it to the new value.
    Otherwise, do nothing.

    :param font_xml: The font XML.
    :param char: A single character.
    :param width: The width of the whitespace, unscaled.
    :param replace: If the character already has data, overwrite.
    :returns: Nothing, font is modified in-place.
    :raises GlyphLimitError: If there are more than 65,535 glyphs in the font; can't add more.
    """
    uni = f"u{ord(char):04X}"
    charstrings = font_xml.find("CFF/CFFFont/CharStrings")
    width *= 125

    if (charstring_xml := font_xml.find(f"{charstrings}/CharString[@name='{uni}']")) is None:
        # Doesn't exist
        lxml.etree.SubElement(charstrings, "CharString", {"name": uni}).text = f"{width} endchar"
        lxml.etree.SubElement(font_xml.find("hmtx"), "mtx", {"name": uni, "width": str(width), "lsb": "0"})
    elif replace:
        # Exists already in the charstrings, replace it too
        charstring_xml.text = f"{width} endchar"
        font_xml.find(f"hmtx/mtx[@name='{uni}']").set("width", str(width))


def add_namerecord_to_font(font_xml: lxml.etree._Element, data: str, name_id: int) -> None:
    """
    Set a namerecord in a font.
    Does not check if such a namerecord already exists, and will add new namerecords.

    * 0 -> Copyright
    * 1 -> Font family
    * 2 -> Font subfamily
    * 3 -> Unique font ID
    * 4 -> Full font name, ID 1 + 2
    * 5 -> Version: "Version maj.min"
    * 6 -> PostScript name
    * 7 -> Trademark
    * 8 -> Manufacturer
    * 9 -> Designer
    * 10 -> Descriptions

    See more at https://docs.microsoft.com/en-us/typography/opentype/spec/name#name-ids.

    All are encoded at (0, 4) and (3, 1), Unicode 2.0+ full.

    :param font_xml: The font XML.
    :param data: Whatever string to add.
    :param name_id: The ID of the namerecord.
    :returns: Nothing, font is modified in-place.
    """
    lxml.etree.SubElement(
        font_xml.find("name"),
        "namerecord",
        {"platformID": "0", "platEncID": "4", "nameID": str(name_id), "langID": "0x0"},
    ).text = data
    lxml.etree.SubElement(
        font_xml.find("name"),
        "namerecord",
        {"platformID": "3", "platEncID": "1", "nameID": str(name_id), "langID": "0x409"},
    ).text = data


def set_cfffont_name(font_xml: lxml.etree._Element, font_name: str, family_name: str) -> None:
    """
    Sets the font's appropriate CFF name in 'CFF ' tables.
    This is not the same as changing the 'name' tables, see :func:`add_namerecord_to_font` instead.

    :param font_xml: The font XML.
    :param font_name: The name of the font (do not include "Regular" or related weights).
    :param family_name: The family name of the font (do not include "Regular" or related weights).
    :returns: Nothing, font is modified in-place.
    """
    cfffont = font_xml.find("CFF/CFFFont")
    cfffont.set("name", mcfonts.utils.sanitize_font_name(font_name))
    cfffont.find("FullName").set("value", font_name)
    cfffont.find("FamilyName").set("value", family_name)


def set_font_times(font_xml: lxml.etree._Element) -> None:
    """
    Sets the font's created and modified times,
    in the format of ``%a %b %d %X %Y``.

    :param font_xml: The font XML.
    :returns: Nothing, font is modified in-place.
    """
    time = datetime.datetime.now().strftime("%a %b %d %X %Y")
    font_xml.find("head/created").set("value", time)
    font_xml.find("head/modified").set("value", time)


def set_notdef_in_font(font_xml: lxml.etree._Element, program: list[str]) -> None:
    """
    Set the .notdef character of the font.
    The font already has a default notdef, use for setting to something else.

    :param font_xml: The font XML.
    :param program: A list of strings of the glyph's program.
    :returns: Nothing, font is modified in-place.
    """
    font_xml.find("CFF/CFFFont/CharStrings/CharString[@name='.notdef']").text = " ".join(str(x) for x in program)
    font_xml.find("hmtx/mtx[@name='.notdef']").set("width", str(program[0]))


def set_font_name(font_xml: lxml.etree._Element, font_name: str, include_credits: bool = True) -> None:
    """
    Set the font's name to `font_xml` in all appropriate places.

    This not the same as :func:`set_cfffont_name`, which sets the name for only CFF tables.

    :param font_xml: The font XML.
    :param font_name: The name of the font.
    :param include_credits: If credits and links to mcfonts are included in the namerecords.
    """
    mcfonts.exporting.add_namerecord_to_font(font_xml, font_name, 1)
    mcfonts.exporting.set_cfffont_name(font_xml, font_name, font_name)
    mcfonts.exporting.add_namerecord_to_font(font_xml, "Regular", 2)
    if include_credits:
        mcfonts.exporting.add_namerecord_to_font(font_xml, "🄯 Public domain", 0)
        mcfonts.exporting.add_namerecord_to_font(font_xml, f"mcfonts: {font_name} Regular: {hash(font_xml)}", 3)
        mcfonts.exporting.add_namerecord_to_font(font_xml, "mcfonts", 8)
        mcfonts.exporting.add_namerecord_to_font(font_xml, f"{font_name} has no trademark.", 7)
        mcfonts.exporting.add_namerecord_to_font(font_xml, "https://gitlab.com/whoatemybutter/mcfonts", 11)
        mcfonts.exporting.add_namerecord_to_font(
            font_xml,
            "This font is under no explicit licensing. "
            "This font's author may have additional licensing terms, contact them.",
            13,
        )
    else:
        mcfonts.exporting.add_namerecord_to_font(font_xml, f"{font_name} Regular: {hash(font_xml)}", 3)
    mcfonts.exporting.add_namerecord_to_font(font_xml, f"{font_name} Regular", 4)
    mcfonts.exporting.add_namerecord_to_font(
        font_xml,
        f"Version 1.0; {datetime.date.strftime(datetime.date.today(), '%B %d, %Y')}",
        5,
    )
    mcfonts.exporting.add_namerecord_to_font(font_xml, mcfonts.utils.sanitize_font_name(font_name), 6)
    mcfonts.exporting.set_font_times(font_xml)



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""General exceptions."""


class GlyphLimitError(Exception):
    """
    Raised when a font has 65,535 characters allocated and that no more can be added.
    """


class RangestringError(Exception):
    """
    Raised when a rangestring is invalid.
    """


class ProviderError(Exception):
    """
    Raised when a :class:`mcfonts.providers.AnyProvider` is expected but not found.
    """



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Constants for mcfonts.

This contains templates for providers, charlists, etc.
Old versions are not stored, and only the latest release is included.
As of this release, 1.19.3 is the latest.

Use :const:`RELEASE_DEFAULT` when looking to template a Vanilla resource without a provider.
"""
import re

import lxml.etree

XML_FONT_TEMPLATE: lxml.etree._Element = lxml.etree.XML(
    b"""<?xml version="1.0" encoding="UTF-8"?>
<ttFont sfntVersion="OTTO" ttLibVersion="4.34">
<GlyphOrder>
<GlyphID id="0" name=".notdef"/>
</GlyphOrder>
<head>
<tableVersion value="1.0"/>
<fontRevision value="1.0"/>
<checkSumAdjustment value="0x910274da"/>
<magicNumber value="0x5f0f3cf5"/>
<flags value="00000000 00001011"/>
<unitsPerEm value="1000"/>
<created value="Mon Jan  1 00:00:00 0000"/>
<modified value="Mon Jan  1 00:00:00 0000"/>
<xMin value="-12500"/>
<yMin value="-12500"/>
<xMax value="12500"/>
<yMax value="12500"/>
<macStyle value="00000000 00000000"/>
<lowestRecPPEM value="8"/>
<fontDirectionHint value="0"/>
<indexToLocFormat value="0"/>
<glyphDataFormat value="0"/>
</head>
<hhea>
<tableVersion value="0x00010000"/>
<ascent value="1000"/>
<descent value="0"/>
<lineGap value="250"/>
<advanceWidthMax value="12500"/>
<minLeftSideBearing value="0"/>
<minRightSideBearing value="0"/>
<xMaxExtent value="1000"/>
<caretSlopeRise value="1"/>
<caretSlopeRun value="0"/>
<caretOffset value="0"/>
<reserved0 value="0"/>
<reserved1 value="0"/>
<reserved2 value="0"/>
<reserved3 value="0"/>
<metricDataFormat value="0"/>
<numberOfHMetrics value="1"/>
</hhea>
<maxp>
<tableVersion value="0x5000"/>
<numGlyphs value="0"/>
</maxp>
<OS_2>
<!-- The fields 'usFirstCharIndex' and 'usLastCharIndex'
will be recalculated by the compiler -->
<version value="4"/>
<xAvgCharWidth value="660"/>
<usWeightClass value="400"/>
<usWidthClass value="5"/>
<fsType value="00000000 00000000"/>
<ySubscriptXSize value="650"/>
<ySubscriptYSize value="700"/>
<ySubscriptXOffset value="0"/>
<ySubscriptYOffset value="140"/>
<ySuperscriptXSize value="650"/>
<ySuperscriptYSize value="700"/>
<ySuperscriptXOffset value="0"/>
<ySuperscriptYOffset value="480"/>
<yStrikeoutSize value="50"/>
<yStrikeoutPosition value="256"/>
<sFamilyClass value="0"/>
<panose>
<bFamilyType value="2"/>
<bSerifStyle value="0"/>
<bWeight value="5"/>
<bProportion value="9"/>
<bContrast value="0"/>
<bStrokeVariation value="0"/>
<bArmStyle value="0"/>
<bLetterForm value="0"/>
<bMidline value="10"/>
<bXHeight value="10"/>
</panose>
<ulUnicodeRange1 value="11111111 11111111 11111111 11111111"/>
<ulUnicodeRange2 value="11111111 11111111 11111111 11111111"/>
<ulUnicodeRange3 value="11111111 11111111 11111111 11111111"/>
<ulUnicodeRange4 value="11111111 11111111 11111111 11111111"/>
<achVendID value="pMCF"/>
<fsSelection value="00000000 10000000"/>
<usFirstCharIndex value="32"/>
<usLastCharIndex value="0"/>
<sTypoAscender value="1000"/>
<sTypoDescender value="0"/>
<sTypoLineGap value="4"/>
<usWinAscent value="1000"/>
<usWinDescent value="0"/>
<ulCodePageRange1 value="00000000 00000000 00000000 00000001"/>
<ulCodePageRange2 value="00000000 00000000 00000000 00000000"/>
<sxHeight value="500"/>
<sCapHeight value="1000"/>
<usDefaultChar value="32"/>
<usBreakChar value="32"/>
<usMaxContext value="1"/>
</OS_2>
<name>
</name>
<cmap>
<tableVersion version="0"/>
</cmap>
<post>
<formatType value="3.0"/>
<italicAngle value="0.0"/>
<underlinePosition value="0"/>
<underlineThickness value="0"/>
<isFixedPitch value="0"/>
<minMemType42 value="0"/>
<maxMemType42 value="0"/>
<minMemType1 value="0"/>
<maxMemType1 value="0"/>
</post>
<CFF>
<major value="1"/>
<minor value="0"/>
<CFFFont name="Default">
<version value="001.000"/>
<Notice value=""/>
<FullName value="Default"/>
<FamilyName value="Default"/>
<Weight value="Regular"/>
<isFixedPitch value="0"/>
<ItalicAngle value="0"/>
<UnderlinePosition value="0"/>
<UnderlineThickness value="0"/>
<PaintType value="0"/>
<CharstringType value="2"/>
<FontMatrix value="0.001 0 0 0.001 0 0"/>
<FontBBox value="0 -1000 1000 1000"/>
<StrokeWidth value="0"/>
<!-- charset is dumped separately as the 'GlyphOrder' element -->
<Encoding name="StandardEncoding"/>
<Private>
<BlueValues value="1 1 16 16"/>
<BlueScale value="0.039625"/>
<BlueShift value="0"/>
<BlueFuzz value="1"/>
<ForceBold value="0"/>
<LanguageGroup value="0"/>
<ExpansionFactor value="0.06"/>
<initialRandomSeed value="0"/>
<defaultWidthX value="750"/>
<nominalWidthX value="0"/>
</Private>
<CharStrings>
<CharString name=".notdef">
750 0 vmoveto 1000 625 -1000 vlineto -125 125 rmoveto 750 -375 -750 vlineto endchar
</CharString>
</CharStrings>
</CFFFont>
<GlobalSubrs/>
</CFF>
<GDEF>
<Version value="0x00010000"/>
<GlyphClassDef/>
</GDEF>
<hmtx>
<mtx name=".notdef" width="750" lsb="0"/>
</hmtx>
</ttFont>"""
)
"""A blank TTF XML, with a ``.notdef`` glyph already embedded."""

EMPTY_FONT_JSON = {"providers": []}
"""An empty font JSON with no providers."""

RELEASE_DEFAULT = {
    "providers": [
        {"type": "space", "advances": {" ": 4, "\u200c": 0}},
        {
            "type": "bitmap",
            "file": "minecraft:font/nonlatin_european.png",
            "ascent": 7,
            "chars": [
                "¡‰­·₴≠¿×ØÞһðøþΑΒ",
                "ΓΔΕΖΗΘΙΚΛΜΝΞΟΠΡΣ",
                "ΤΥΦΧΨΩαβγδεζηθικ",
                "λμνξοπρςστυφχψωЂ",
                "ЅІЈЉЊЋАБВГДЕЖЗИК",
                "ЛМНОПРСТУФХЦЧШЩЪ",
                "ЫЬЭЮЯабвгдежзикл",
                "мнопрстуфхцчшщъы",
                "ьэюяєѕіјљњ–—‘’“”",
                "„…⁊←↑→↓⇄＋ƏəɛɪҮүӨ",
                "өʻˌ;ĸẞß₽€ѢѣѴѵӀѲѳ",
                "⁰¹³⁴⁵⁶⁷⁸⁹⁺⁻⁼⁽⁾ⁱ™",
                "ʔʕ⧈⚔☠ҚқҒғҰұӘәҖҗҢ",
                "ңҺאבגדהוזחטיכלמם",
                "נןסעפףצץקר¢¤¥©®µ",
                "¶¼½¾·‐‚†‡•‱′″‴‵‶",
                "‷‹›※‼‽⁂⁈⁉⁋⁎⁏⁑⁒⁗℗",
                "−∓∞☀☁☈Є☲☵☽♀♂⚥♠♣♥",
                "♦♩♪♫♬♭♮♯⚀⚁⚂⚃⚄⚅ʬ⚡",
                "⛏✔❄❌❤⭐⸘⸮⸵⸸⹁⹋⥝ᘔƐ߈",
                "ϛㄥⱯᗺƆᗡƎℲ⅁ꞰꞀԀꝹᴚ⟘∩",
                "Ʌ⅄ɐɔǝɟᵷɥᴉɾʞꞁɯɹʇʌ",
                "ʍʎԱԲԳԴԶԷԹԺԻԼԽԾԿՀ",
                "ՁՂՃՄՅՆՇՈՉՋՌՍՎՏՐՑ",
                "ՒՓՔՕՖՙաբգդեզէըթժ",
                "իլխծկհձղճմյնշոչպ",
                "ջռսվտրցւփքօֆևשתԸ",
                "՚՛՜՝՞՟ՠֈ֏¯ſƷʒǷƿȜ",
                "ȝȤȥ˙Ꝛꝛ‑⅋⏏⏩⏪⏭⏮⏯⏴⏵",
                "⏶⏷⏸⏹⏺⏻⏼⏽⭘▲▶▼◀●◦◘",
                "⚓⛨ĲĳǉꜨꜩꜹꜻﬀﬁﬂﬃﬅ�Ե",
                "Պᚠᚢᚣᚤᚥᚦᚧᚨᚩᚪᚫᚬᚭᚮᚯ",
                "ᚰᚱᚲᚳᚴᚶᚷᚸᚹᚺᚻᚼᚽᚾᚿᛀ",
                "ᛁᛂᛃᛄᛅᛆᛇᛈᛉᛊᛋᛌᛍᛎᛏᛐ",
                "ᛑᛒᛓᛔᛕᛖᛗᛘᛙᛚᛛᛜᛝᛞᛟᛠ",
                "ᛡᛢᛣᛤᛥᛦᛧᛨᛩᛪ᛫᛬᛭ᛮᛯᛰ",
                "ᛱᛲᛳᛴᛵᛶᛷᛸ☺☻¦☹ך׳״װ",
                "ױײ־׃׆´¨ᴀʙᴄᴅᴇꜰɢʜᴊ",
                "ᴋʟᴍɴᴏᴘꞯʀꜱᴛᴜᴠᴡʏᴢ§",
                "ɱɳɲʈɖɡʡɕʑɸʝʢɻʁɦʋ",
                "ɰɬɮʘǀǃǂǁɓɗᶑʄɠʛɧɫ",
                "ɨʉʊɘɵɤɜɞɑɒɚɝƁƉƑƩ",
                "ƲႠႡႢႣႤႥႦႧႨႩႪႫႬႭႮ",
                "ႯႰႱႲႳႴႵႶႷႸႹႺႻႼႽႾ",
                "ႿჀჁჂჃჄჅჇჍაბგდევზ",
                "თიკლმნოპჟრსტუფქღ",
                "ყშჩცძწჭხჯჰჱჲჳჴჵჶ",
                "ჷჸჹჺ჻ჼჽჾჿתּשׂפֿפּכּײַיִ",
                "וֹוּבֿבּꜧꜦɺⱱʠʗʖɭɷɿʅʆ",
                "ʓʚ₪₾֊ⴀⴁⴂⴃⴄⴅⴆⴡⴇⴈⴉ",
                "ⴊⴋⴌⴢⴍⴎⴏⴐⴑⴒⴣⴓⴔⴕⴖⴗ",
                "ⴘⴙⴚⴛⴜⴝⴞⴤⴟⴠⴥ⅛⅜⅝⅞⅓",
                "⅔✉☂☔☄⛄☃⌛⌚⚐✎❣♤♧♡♢",
                "⛈☰☱☳☴☶☷↔⇒⇏⇔⇵∀∃∄∉",
                "∋∌⊂⊃⊄⊅∧∨⊻⊼⊽∥≢⋆∑⊤",
                "⊥⊢⊨≔∁∴∵∛∜∂⋃⊆⊇□△▷",
                "▽◁◆◇○◎☆★✘₀₁₂₃₄₅₆",
                "₇₈₉₊₋₌₍₎∫∮∝⌀⌂⌘〒ɼ",
                "ƄƅẟȽƚƛȠƞƟƧƨƪƸƹƻƼ",
                "ƽƾȡȴȵȶȺⱥȻȼɆɇȾⱦɁɂ",
                "ɃɄɈɉɊɋɌɍɎɏẜẝỼỽỾỿ",
                "Ꞩꞩ𐌰𐌱𐌲𐌳𐌴𐌵𐌶𐌷𐌸𐌹𐌺𐌻𐌼𐌽",
                "𐌾𐌿𐍀𐍁𐍂𐍃𐍄𐍅𐍆𐍇𐍈𐍉𐍊🌧🔥🌊",
                "⅐⅑⅕⅖⅗⅙⅚⅟↉🗡🏹🪓🔱🎣🧪⚗",
                "⯪⯫Ɑ🛡✂🍖🪣🔔⏳⚑₠₡₢₣₤₥",
                "₦₩₫₭₮₰₱₲₳₵₶₷₸₹₺₻",
                "₼₿              ",
            ],
        },
        {
            "type": "bitmap",
            "file": "minecraft:font/accented.png",
            "height": 12,
            "ascent": 10,
            "chars": [
                "ÀÁÂÃÄÅÆÇÈÉÊËÌÍÎÏ",
                "ÐÑÒÓÔÕÖÙÚÛÜÝàáâã",
                "äåæçìíîïñòóôõöùú",
                "ûüýÿĀāĂăĄąĆćĈĉĊċ",
                "ČčĎďĐđĒēĔĕĖėĘęĚě",
                "ĜĝḠḡĞğĠġĢģĤĥĦħĨĩ",
                "ĪīĬĭĮįİıĴĵĶķĹĺĻļ",
                "ĽľĿŀŁłŃńŅņŇňŊŋŌō",
                "ŎŏŐőŒœŔŕŖŗŘřŚśŜŝ",
                "ŞşŠšŢţŤťŦŧŨũŪūŬŭ",
                "ŮůŰűŲųŴŵŶŷŸŹźŻżŽ",
                "žǼǽǾǿȘșȚțΆΈΉΊΌΎΏ",
                "ΐΪΫάέήίΰϊϋόύώЀЁЃ",
                "ЇЌЍЎЙйѐёђѓїћќѝўџ",
                "ҐґḂḃḊḋḞḟḢḣḰḱṀṁṖṗ",
                "ṠṡṪṫẀẁẂẃẄẅỲỳèéêë",
                "ŉǧǫЏḍḥṛṭẒỊịỌọỤụ№",
                "ȇƔɣʃ⁇ǱǲǳǄǅǆǇǈǊǋǌ",
                "ℹᵫꜲꜳꜴꜵꜶꜷꜸꜺꜼꜽꝎꝏꝠꝡ",
                "ﬄﬆᚡᚵƠơƯưẮắẤấẾếốỚ",
                "ớỨứẰằẦầỀềồỜờỪừẢả",
                "ẲẳẨẩẺẻổỞỂểỈỉỎỏỔở",
                "ỦủỬửỶỷẠạẶặẬậẸẹỆệ",
                "ỘộỢợỰựỴỵỐƕẪẫỖỗữ☞",
                "☜☮ẴẵẼẽỄễỒỠỡỮỸỹҘҙ",
                "ҠҡҪҫǶ⚠⓪①②③④⑤⑥⑦⑧⑨",
                "⑩⑪⑫⑬⑭⑮⑯⑰⑱⑲⑳ⒶⒷⒸⒹⒺ",
                "ⒻⒼⒽⒾⒿⓀⓁⓂⓃⓄⓅⓆⓇⓈⓉⓊ",
                "ⓋⓌⓍⓎⓏⓐⓑⓒⓓⓔⓕⓖⓗⓘⓙⓚ",
                "ⓛⓜⓝⓞⓟⓠⓡⓢⓣⓤⓥⓦⓧⓨⓩ̧",
                "ʂʐɶǍǎǞǟǺǻȂȃȦȧǠǡḀ",
                "ḁȀȁḆḇḄḅᵬḈḉḐḑḒḓḎḏ",
                "ḌᵭḔḕḖḗḘḙḜḝȨȩḚḛȄȅ",
                "ȆᵮǴǵǦḦḧḨḩḪḫȞȟḤẖḮ",
                "ḯȊȋǏǐȈȉḬḭǰȷǨǩḲḳḴ",
                "ḵḺḻḼḽḶḷḸḹⱢḾḿṂṃᵯṄ",
                "ṅṆṇṊṋǸǹṈṉᵰǬǭȬȭṌṍ",
                "ṎṏṐṑṒṓȎȏȪȫǑǒȮȯȰȱ",
                "ȌȍǪṔṕᵱȒȓṘṙṜṝṞṟȐȑ",
                "ṚᵳᵲṤṥṦṧṢṣṨṩᵴṰṱṮṯ",
                "ṬẗᵵṲṳṶṷṸṹṺṻǓǔǕǖǗ",
                "ǘǙǚǛǜṴṵȔȕȖṾṿṼṽẆẇ",
                "ẈẉẘẌẍẊẋȲȳẎẏẙẔẕẐẑ",
                "ẓᵶǮǯẛꜾꜿǢǣᵺỻᴂᴔꭣȸʣ",
                "ʥʤʩʪʫȹʨʦʧꭐꭑ₧Ỻאַאָƀ",
                "ƂƃƇƈƊƋƌƓǤǥƗƖɩƘƙƝ",
                "ƤƥɽƦƬƭƫƮȗƱƜƳƴƵƶƢ",
                "ƣȢȣʭʮʯﬔﬕﬗﬖﬓӐӑӒӓӶ",
                "ӷҔҕӖӗҼҽҾҿӚӛӜӝӁӂӞ",
                "ӟӢӣӤӥӦӧӪӫӰӱӮӯӲӳӴ",
                "ӵӸӹӬӭѶѷӔӺԂꚂꚀꚈԪԬꚄ",
                "ԄԐӠԆҊӃҞҜԞԚӅԮԒԠԈԔ",
                "ӍӉԨӇҤԢԊҨԤҦҎԖԌꚐҬꚊ",
                "ꚌԎҲӼӾԦꚔҴꚎҶӋҸꚒꚖꚆҌ",
                "ԘԜӕӻԃꚃꚁꚉԫԭꚅԅԑӡԇҋ",
                "ӄҟҝԟԛӆԯԓԡԉԕӎӊԩӈҥ",
                "ԣԋҩԥҧҏԗԍꚑҭꚋꚍԏҳӽӿ",
                "ԧꚕҵꚏҷӌҹꚓꚗꚇҍԙԝἈἀἉ",
                "ἁἊἂἋἃἌἄἍἅἎἆἏἇᾺὰᾸ",
                "ᾰᾹᾱΆάᾈᾀᾉᾁᾊᾂᾋᾃᾌᾄᾍ",
                "ᾅᾎᾆᾏᾇᾼᾴᾶᾷᾲᾳἘἐἙἑἚ",
                "ἒἛἓἜἔἝἕῈΈὲέἨἠῊὴἩ",
                "ἡἪἢἫἣἬἤἭἥἮἦἯἧᾘᾐᾙ",
                "ᾑᾚᾒᾛᾓᾜᾔᾝᾕᾞᾖᾟᾗΉήῌ",
                "ῃῂῄῆῇῚὶΊίἸἰἹἱἺἲἻ",
                "ἳἼἴἽἵἾἶἿἷῘῐῙῑῒΐῖ",
                "ῗῸὸΌόὈὀὉὁὊὂὋὃὌὄὍ",
                "ὅῬῤῥῪὺΎύὙὑὛὓὝὕὟὗ",
                "ῨῠῩῡϓϔῢΰῧὐὒὔῦὖῺὼ",
                "ΏώὨὠὩὡὪὢὫὣὬὤὭὥὮὦ",
                "Ὧὧᾨᾠᾩᾡᾪᾢᾫᾣᾬᾤᾭᾥᾮᾦ",
                "ᾯᾧῼῳῲῴῶῷ☯☐☑☒ƍƺⱾȿ",
                "ⱿɀᶀꟄꞔᶁᶂᶃꞕᶄᶅᶆᶇᶈᶉᶊ",
                "ᶋᶌᶍꟆᶎᶏᶐᶒᶓᶔᶕᶖᶗᶘᶙᶚ",
                "ẚ⅒⅘₨₯           ",
            ],
        },
        {
            "type": "bitmap",
            "file": "minecraft:font/ascii.png",
            "ascent": 7,
            "chars": [
                "                ",
                "                ",
                " !\"#$%&'()*+,-./",
                "0123456789:;<=>?",
                "@ABCDEFGHIJKLMNO",
                "PQRSTUVWXYZ[\\]^_",
                "`abcdefghijklmno",
                "pqrstuvwxyz{|}~ ",
                "                ",
                "            £  ƒ",
                "      ªº  ¬   «»",
                "░▒▓│┤╡╢╖╕╣║╗╝╜╛┐",
                "└┴┬├─┼╞╟╚╔╩╦╠═╬╧",
                "╨╤╥╙╘╒╓╫╪┘┌█▄▌▐▀",
                "             ∅∈ ",
                "≡±≥≤⌠⌡÷≈°∙ √ⁿ²■ ",
            ],
        },
        {
            "type": "legacy_unicode",
            "sizes": "minecraft:font/glyph_sizes.bin",
            "template": "minecraft:font/unicode_page_%s.png",
        },
    ]
}
"""The font JSON of default.json, the default Minecraft font."""
RELEASE_ALT = {
    "providers": [
        {
            "type": "bitmap",
            "file": "minecraft:font/ascii_sga.png",
            "ascent": 7,
            "chars": [
                "                ",
                "                ",
                "                ",
                "                ",
                " ABCDEFGHIJKLMNO",
                "PQRSTUVWXYZ     ",
                " abcdefghijklmno",
                "pqrstuvwxyz     ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
                "                ",
            ],
        }
    ]
}
"""The font JSON of alt.json, the SGA font."""
RELEASE_UNIFORM = {
    "providers": [
        {
            "type": "legacy_unicode",
            "sizes": "minecraft:font/glyph_sizes.bin",
            "template": "minecraft:font/unicode_page_%s.png",
        }
    ]
}
"""The font JSON of uniform.json, the Unicode fallback font."""

PROVIDER_NONLATIN = RELEASE_DEFAULT["providers"][1]
"""The default provider for nonlatin-european."""
PROVIDER_ACCENTED = RELEASE_DEFAULT["providers"][2]
"""The default provider for accented."""
PROVIDER_ASCII = RELEASE_DEFAULT["providers"][3]
"""The default provider for ASCII."""

CHARLIST_NONLATIN = PROVIDER_NONLATIN["chars"]
"""The default charlist for nonlatin-european."""
CHARLIST_ACCENTED = PROVIDER_ACCENTED["chars"]
"""The default charlist for accented."""
CHARLIST_ASCII = PROVIDER_ASCII["chars"]
"""The default charlist for ASCII."""

MAX_I32 = 4294967295


RANGESTRING_PART = re.compile(r"!?(((0x|U\+|&#x)([0-9A-F]{1,7});?)|(.))", re.I)
RANGESTRING_RANGE_DELIMITER = re.compile(r"(-|\.\.)")
RANGESTRING_RANGE = re.compile(
    f"({RANGESTRING_PART.pattern}{RANGESTRING_RANGE_DELIMITER.pattern}{RANGESTRING_PART.pattern})",
    re.I,
)
RANGESTRING_COMPONENT = re.compile(f"({RANGESTRING_RANGE.pattern}|{RANGESTRING_PART.pattern})", re.I)
RANGESTRING_COMPONENT_DELIMITER = re.compile("(, ?)")
RANGESTRING_FULL = re.compile(f"(({RANGESTRING_COMPONENT.pattern}{RANGESTRING_COMPONENT_DELIMITER.pattern})*)", re.I)

SCHEMA_PATTERN_NAMESPACE = re.compile(r"^(?:[0-9a-z_\-.]:)?.*$")

SCHEMA_PROVIDER_BITMAP = {
    "type": "object",
    "properties": {
        "type": {"const": "bitmap"},
        "file": {"type": "string", "minLength": 1, "pattern": SCHEMA_PATTERN_NAMESPACE.pattern},
        "height": {"type": "integer", "minimum": 1, "maximum": MAX_I32, "default": 8},
        "ascent": {"type": "integer", "minimum": -MAX_I32, "maximum": MAX_I32},
        "chars": {"type": "array", "minItems": 1, "items": {"type": "string"}},
        "comment": {"type": "string"},
    },
    "required": ["type", "file", "ascent", "chars"],
    "additionalProperties": False,
}
"""A JSON schema to validate "bitmap" providers against."""

SCHEMA_PROVIDER_SPACE = {
    "type": "object",
    "properties": {
        "type": {"const": "space"},
        "advances": {
            "type": "object",
            "minLength": 1,
            "patternProperties": {RANGESTRING_FULL.pattern: {"type": "integer", "minimum": -256, "maximum": 256}},
        },
        "comment": {"type": "string"},
    },
    "required": ["type", "advances"],
    "additionalProperties": False,
}
"""A schema to validate "space" providers against."""

SCHEMA_PROVIDER_LEGACY_UNICODE = {
    "type": "object",
    "properties": {
        "type": {"const": "legacy_unicode"},
        "sizes": {"type": "string", "minLength": 1, "pattern": SCHEMA_PATTERN_NAMESPACE.pattern},
        "template": {"type": "string", "minLength": 1, "pattern": SCHEMA_PATTERN_NAMESPACE.pattern},
        "comment": {"type": "string"},
    },
    "required": ["type", "sizes", "template"],
    "additionalProperties": False,
}
"""A schema to validate "legacy_unicode" providers against."""

SCHEMA_PROVIDER_TTF = {
    "type": "object",
    "properties": {
        "type": {"const": "ttf"},
        "file": {"type": "string", "minLength": 1, "pattern": SCHEMA_PATTERN_NAMESPACE.pattern},
        "shift": {"type": "array", "items": {"type": "number"}, "minItems": 2, "maxItems": 2},
        "size": {"type": "number", "minimum": 1},
        "oversample": {"type": "number", "minimum": 1},
        "skip": {"type": ["string", "array"]},
        "comment": {"type": "string"},
    },
    "required": ["type", "file", "shift", "size", "oversample"],
    "additionalProperties": False,
}
"""A schema to validate "ttf" providers against."""

SCHEMA_PROVIDER_OPTIONS = {
    "type": "object",
    "properties": {
        "type": {"enum": ["options", "mcfonts:options"]},
        "fields": {
            "type": "object",
            "properties": {
                "fullwidth": {
                    "type": "object",
                    "patternProperties": {RANGESTRING_FULL.pattern: {"type": "boolean"}},
                },
                "shift": {
                    "type": "object",
                    "patternProperties": {
                        RANGESTRING_FULL.pattern: {
                            "type": "array",
                            "minItems": 2,
                            "maxItems": 2,
                            "items": {"type": "integer"},
                        }
                    },
                },
                "spacing": {
                    "type": "object",
                    "patternProperties": {RANGESTRING_FULL.pattern: {"type": "integer", "minimum": 0}},
                },
                "width": {
                    "type": "object",
                    "patternProperties": {RANGESTRING_FULL.pattern: {"type": "integer"}},
                },
            },
        },
        "comment": {"type": "string"},
    },
    "required": ["type"],
    "additionalProperties": False,
}
"""A schema to validate "options" providers against."""

OPTION_FIELDS = {
    "width",
    "fullwidth",
    "shift",
    "spacing",
}

PADDING_CHARS = {"\0", " "}
"""Characters that act as padding; glyphs cannot be assigned to these chars."""

UNKNOWN_FIELD = "???"



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Utilities for the "bitmap" font provider.
Most of these are for either aligning sheets, getting glyphs, or padding charlists.
"""
import itertools
import typing

import PIL.Image
import PIL.ImageChops

import mcfonts.constants
import mcfonts.exporting
import mcfonts.providers
import mcfonts.unicode
import mcfonts.utils


def pad_charlist(
    charlist: list[str],
    pad_amount: int = 0,
    pad_char: str = " ",
    pad_from: typing.Literal["left", "right"] = "right",
) -> list[str]:
    """
    Pad `charlist` so that all of its strings have the same length.

    :param charlist: Character list to pad.
    :param pad_amount: What to pad the characters to; the ideal width.
        If this is 0, this will be determined from the largest length of strings in `charlist`.
    :param pad_char: The character to pad with, default is a space.
    :param pad_from: Direction to pad the string from, left => "___Text", right => "Text___".
    :returns: The padded charlist list.
    """
    if pad_from not in ("left", "right"):
        raise ValueError('pad_from must be "left" or "right"')
    if pad_amount == 0:
        # Get the largest string based on length
        pad_amount = len(max(charlist, key=len))
    for index, charline in enumerate(charlist):
        if len(charline) != pad_amount:
            if pad_from == "left":
                # Rjust means justify to right
                charlist[index] = charline.rjust(pad_amount, pad_char[0])
            else:
                # Ljust means justify to left
                charlist[index] = charline.ljust(pad_amount, pad_char[0])
    return charlist


def assert_charlist_is_equal(charlist: list[str]) -> tuple[bool, int]:
    """
    Ensure that every string inside `charlist` has the same length as the first string.

    :param charlist: The charlist of strings to check lengths of.
    :returns: A tuple of (if the lengths are equal, the expected length).
    """
    ideal = len(max(charlist, key=len))
    for charline in charlist:
        if len(charline) != ideal:
            return False, ideal
    return True, ideal


def align_font_texture(
    sheet: PIL.Image.Image, glyph_cell: tuple[int, int] = (8, 8), new_left_padding: int = 0
) -> PIL.Image.Image:
    """
    Align a font sheet's character's sheet in order to what `new_left_padding` is.

    This only shifts the character on the X axis (horizontally, left-right).

    :param sheet: The font sheet, not the individual glyph.
    :param glyph_cell: The dimensions of each individual character, their cell bounding box.
    :param new_left_padding: The distance away from the left edge the new character should be.
    :returns: The new font sheet.
    """
    height = sheet.height // glyph_cell[1]
    width = sheet.width // glyph_cell[0]
    for y_position in range(height):
        for x_position in range(width):
            # Glyph dimensions before alignment
            dimensions = (
                x_position * glyph_cell[0],
                y_position * glyph_cell[1],
                (x_position + 1) * glyph_cell[0],
                (y_position + 1) * glyph_cell[1],
            )
            # Get the glyph itself
            glyph = sheet.crop(dimensions)
            # If width is not the new padding, offset it by the difference
            if (margin := mcfonts.exporting.get_glyph_nominals(glyph)[0]) != new_left_padding:
                sheet.paste(PIL.ImageChops.offset(glyph, new_left_padding - margin, 0))
    return sheet


def get_largest_effective_glyph_dimensions(
    glyphs: list[PIL.Image.Image | None], isolate_dimensions: bool = True
) -> tuple[int, int]:
    """
    Given a :class:`collections.deque` of glyphs,
    return the width and height of the largest character in that deque.

    This function operates based on the actual pixels inside an image; it crops out empty space
    and then determines the width and height. For a function that operates on an image's true
    width and height, see :func:`get_largest_glyph_dimensions`.

    :param glyphs: A deque of the glyphs to iterate over.
    :param isolate_dimensions:
        If True, the maximum height and width will be calculated separately.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 9).
        If False, the maximum height and width is dependent on the maximum width found.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 8).
        It is recommended that you leave this to True.
    :returns: A tuple of (maximum width, maximum height).
    """
    max_width = 0
    max_height = 0
    for glyph in glyphs:
        if glyph:
            if bbox := glyph.getbbox():
                if (width := bbox[2] - bbox[0]) > max_width:
                    max_width = width
                    if not isolate_dimensions:
                        max_height = width
                    continue
                if (height := bbox[3] - bbox[1]) > max_height:
                    max_height = height
    return max_width, max_height


def get_largest_glyph_dimensions(
    glyphs: list[PIL.Image.Image | None], isolate_dimensions: bool = True
) -> tuple[int, int]:
    """
    Given a :class:`collections.deque` of glyphs,
    return the width and height of the largest character in that deque.

    This function is different from :func:`get_largest_effective_glyph_dimensions` in that it
    does not calculate the real width and height that pixels in an image extend to. It only
    relies on the width and height of the *canvas* of the image.

    :param glyphs: A deque of the glyphs to iterate over.
    :param isolate_dimensions:
        If True, the maximum height and width will be calculated separately.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 9).

        If False, the maximum height and width is dependent on the maximum width found.
        This means that dimensions (4, 6) (12, 8) (3, 9) will be (12, 8).

        It is recommended that you leave this to True.
    :returns: A tuple of (maximum width, maximum height).
    """
    max_width = 0
    max_height = 0
    for glyph in glyphs:
        if glyph:
            if glyph.getbbox():
                if (width := glyph.width) > max_width:
                    max_width = width
                    if not isolate_dimensions:
                        max_height = width
                    continue
                if (height := glyph.height) > max_height:
                    max_height = height
    return max_width, max_height


def get_glyphs_from_resource(
    resource: PIL.Image.Image, char_counts: tuple[int, int]
) -> typing.Iterator[PIL.Image.Image | None]:
    """
    Given a resource and number of characters in each row and column of that resource,
    yield every :class:`PIL.Image.Image` in that resource.

    If no data exists at a cell, yield None.

    :param resource: A :class:`PIL.Image.Image` of the font resource.
    :param char_counts: A tuple of the number of characters in each row and column.
    :returns: A yield of :class:`PIL.Image.Image` glyphs, or None.
    """
    height = resource.height // char_counts[1]
    width = resource.width // char_counts[0]
    for y_index in range(char_counts[1]):
        for x_index in range(char_counts[0]):
            if mcfonts.utils.is_image_empty(
                glyph := resource.crop(
                    (
                        x_index * width,
                        y_index * height,
                        (x_index + 1) * width,
                        (y_index + 1) * height,
                    )
                )
            ):
                yield None
            else:
                yield glyph


def range_file_to_charlist(resource: PIL.Image.Image, char: str, cell_bound: tuple[int, int] = (8, 8)) -> list[str]:
    """
    Given a path to a resource and a starting character, return an appropriate charlist.

    Return a provider that has a charlist that correctly encompasses all chars covered by
    the resource. Glyphs in the resource that are empty are skipped over.

    :param resource: The resource to grab textures from.
    :param char:
        A single character.
        The codepoint of this will increase by 1 on every glyph.
    :param cell_bound:
        The dimensions of an individual glyph in `resource`.
        Glyph dimensions must be the same throughout the whole of `resource`.
    :returns: A `chars` list.
    """
    charlist: list[str] = []
    charline: list[str] = []
    codepoint = ord(char)
    for height_offset in range(resource.height // cell_bound[1]):
        for width_offset in range(resource.width // cell_bound[0]):
            if mcfonts.utils.is_image_empty(
                resource.crop(
                    (
                        width_offset * cell_bound[0],
                        height_offset * cell_bound[1],
                        (width_offset + 1) * cell_bound[0],
                        (height_offset + 1) * cell_bound[1],
                    )
                )
            ):
                charline.append(" ")
            else:
                charline.append(chr(codepoint))
            codepoint += 1
        charlist.append("".join(charline))
        charline.clear()
    return charlist


def load_resources(
    providers: list[dict],
    json_path: str,
    strict: bool = False,
    mode: str = "LA",
) -> dict[str, PIL.Image.Image | bytes]:
    """
    Look through every provider and load their source files into a dictionary.

    :param providers: A list of the providers to get resources from.
    :param json_path: The path to the original font JSON.
    :param strict:
        If True:
        * Missing files will raise an error

        If False:
        * Missing files will be skipped

    :param mode: Mode to load the images in. See PIL modes.
    :returns: The loaded resources.
    :raises FileNotFoundError: If a referenced file is not found and `strict` is True.
    """
    resources: dict[str, PIL.Image.Image | bytes] = {}
    for provider in providers:
        try:
            if (provider_type := provider.get("type", "")).lower().strip() == "bitmap":
                file = provider.get("file", "")
                resources[file] = load_resources_bitmap(file, json_path).convert(mode)
            elif provider_type == "legacy_unicode":
                resources |= dict(
                    load_resources_legacy_unicode(provider.get("template", ""), json_path, provider.get("sizes", ""))
                )
            elif provider_type == "ttf":
                file = provider.get("file", "")
                resources[file] = load_resources_ttf(file, json_path)
        except PIL.UnidentifiedImageError:
            mcfonts.logger.warning(
                mcfonts.providers.format_provider_message(provider, "has invalid file; must be PNG.")
            )
            continue
        except FileNotFoundError as exception:
            if not strict:
                # Missing files are ignored
                mcfonts.logger.warning(
                    mcfonts.providers.format_provider_message(
                        provider, f'file "{exception.filename}" does not exist, skipping.'
                    )
                )
                continue
            # Missing files not allowed, raising exception
            raise exception
    return resources


def load_resources_bitmap(file_name: str, json_path: str) -> PIL.Image.Image:
    """
    A convenient companion function to :func:`load_resources`,
    but specialized for "bitmap" provider resources.

    :param file_name: The name of the file to search for.
    :param json_path: The path that the original font JSON was located in.
    :return: A :class:`PIL.Image.Image` instance.
    """
    with open(
        mcfonts.utils.resolve_resource_path(file_name, json_path, "textures"),
        "rb",
    ) as open_tempfile:
        image = PIL.Image.open(open_tempfile, formats=["png"])
        image.load()
        return image


def load_resources_legacy_unicode(
    template: str, json_path: str, sizes: str
) -> typing.Iterator[tuple[str, PIL.Image.Image | bytes]]:
    """
    A convenient companion function to :func:`load_resources`,
    but specialized for "legacy_unicode" provider resources.

    :param template: The template string, must include "%s".
    :param json_path: The path that the original font JSON was located in.
    :param sizes: The name of the "sizes" file.
    :returns: A yield of {filename: image}.
    """
    template_parts: list[str] = template.split("%s", 1)
    # Skip surrogates
    for codepage in itertools.chain(range(0xD8), range(0xE0, 0x100)):
        template_name = f"{template_parts[0]}{codepage:02x}{template_parts[1]}"
        try:
            with open(
                mcfonts.utils.resolve_resource_path(template_name, json_path, "font"),
                "rb",
            ) as open_tempfile:
                image = PIL.Image.open(open_tempfile, formats=["png"])
                image.load()
                yield template_name, image
        except FileNotFoundError:
            # Page doesn't exist, acceptable error
            continue
    with open(
        mcfonts.utils.resolve_resource_path(sizes, json_path),
        "rb",
    ) as open_tempfile:
        yield sizes, open_tempfile.read()


def load_resources_ttf(file_name: str, json_path: str) -> bytes:
    """
    A convenient companion function to :func:`load_resources`,
    but specialized for "ttf" provider resources.

    :param file_name: The name of the TTF file to search for.
    :param json_path: The path that the original font JSON was located in.
    :return: Bytes of the TTF file.
    """
    with open(
        mcfonts.utils.resolve_resource_path(file_name, json_path, "font"),
        "rb",
    ) as open_tempfile:
        return open_tempfile.read()


def validate_charlist(
    charlist: list[str],
    resource: PIL.Image.Image | None = None,
    glyph_cell: tuple[int, int] = (8, 8),
):
    """
    Given `charlist`, ensure that it is valid and does not have any possible issues.
    Issues include:

    * Empty lines
    * Uneven lines
    * Duplicate characters
    * \\*Padding chars in places where glyph exists in `resource`
    * \\*Characters that will have an empty glyph in `resource`

    \\* Optional, requires `resource`.

    :param charlist: A charlist to validate.
    :param resource: An optional :class:`PIL.Image.Image` to validate glyph textures against.
    :param glyph_cell: Tuple of (glyph width, glyph height) for each glyph in `resource`.
    """
    char_traces: dict[str, tuple[int, int]] = {}
    for y_index, charline in enumerate(charlist):
        if not (len_charline := len(charline)):
            mcfonts.logger.warning(f"Charline {y_index} is empty.")
        elif not (charlist_length := mcfonts.utils.bitmap.assert_charlist_is_equal(charlist)):
            mcfonts.logger.warning(
                "Uneven amount of characters in charlist,"
                f"expected {charlist_length}, got {len_charline} on line {y_index}.",
            )
        for x_index, character in enumerate(charline):
            if character not in mcfonts.constants.PADDING_CHARS and character in char_traces:
                trace = char_traces[character]
                mcfonts.logger.warning(
                    f"Duplicate character {mcfonts.utils.pretty_print_char(character)} "
                    f"on charline {y_index} index {x_index}. "
                    f"Character was already defined at charline {trace[0]} index {trace[1]}.",
                )
            else:
                char_traces[character] = (y_index, x_index)
            if resource:
                is_glyph_invisible = mcfonts.utils.is_image_invisible(
                    resource.crop(
                        (
                            x_index * glyph_cell[0],
                            y_index * glyph_cell[1],
                            (x_index + 1) * glyph_cell[0],
                            (y_index + 1) * glyph_cell[1],
                        )
                    )
                )
                if not is_glyph_invisible and character in mcfonts.constants.PADDING_CHARS:
                    mcfonts.logger.debug(
                        "Padding character on " f"charline {y_index} index {x_index}, but glyph data exists.",
                    )
                elif is_glyph_invisible and not mcfonts.unicode.is_char_invisible(character):
                    mcfonts.logger.debug(
                        "Empty glyph for character "
                        f"{mcfonts.utils.pretty_print_char(character)} on "
                        f"charline {y_index} index {x_index}.",
                    )



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Utilities for Minecraft fonts. Not all of these are used.
"""
import os
import re
import typing
import unicodedata

import PIL.Image
import tinyunicodeblock
import mcfonts.colors
import mcfonts.constants
import mcfonts.unicode


def resolve_resource_path(path: str, json_path: str, subpath: str = "") -> str:
    """
    Resolve the correct path to the resource indicated by ``path``,
    using ``json_path`` as the base path to follow from.
    ``json_path`` is the font's JSON path, ``path`` is the request resource path.

    >>> mcfonts.utils.resolve_resource_path("mypath", "mypath/font.json", "textures")
    '~/mcfonts/textures/mypath'

    :param path: Unresolved path to the resource.
    :param json_path: Path to the font JSON.
    :param subpath: Resource will be loaded from the ``subpath`` directory.
    :returns: An absolute path of the request resource, ``path``.
    """
    temppath = path.split(":", 1)
    if len(temppath) > 1:
        # Uses a namespace
        return expand_path(
            os.path.join(
                json_path,
                f"../../../{temppath[0]}/{subpath}/{temppath[1]}",
            )
        )
    return expand_path(os.path.join(json_path, f"../../{subpath}/{temppath[0]}"))


def pretty_print_char(char: str) -> str:
    """
    Put relevant about a character into a string,
    following ``U+<codepoint> <name> <character>``.

    >>> mcfonts.utils.pretty_print_char('\u2601')
    'U+2601: CLOUD ☁'
    >>> mcfonts.utils.pretty_print_char('\ue000')
    'U+E000: <PRIVATE USE> \ue000'
    >>> mcfonts.utils.pretty_print_char('\U0001f400')
    'U+1F400: RAT 🐀'
    >>> mcfonts.utils.pretty_print_char('\b')
    'U+0008: BACKSPACE ␈'

    :param char: A single character.
    :returns: The pretty character string.
    """
    codepoint = ord(char)
    if unicodedata.category(char) == "Cc":
        return (
            f"U+{codepoint:04X}: "
            f"{unicodedata.name(chr(ord(char) + 0x2400)).split('SYMBOL FOR ')[1]} "
            f"{chr(ord(char) + 0x2400)}"
        )
    try:
        return f"U+{codepoint:04X}: {unicodedata.name(char)} {char}"
    except ValueError:
        if mcfonts.unicode.is_char_private_use(char):
            return f"U+{codepoint:04X}: <PRIVATE USE> {char}"
        return f"U+{codepoint:04X}: {char}"


def sanitize_font_name(font_name: str) -> str:
    """
    Ensure ``sanitized_font_name`` is a valid PostScript font name.
    A PostScript font name cannot:

    * Contain ``(){}[]<;>%/`` or space
    * Be longer than 63 characters
    * Have non-ASCII characters

    >>> mcfonts.utils.sanitize_font_name("\u2600 This is a really long name!")
    '_This_is_a_really_long_name!'

    :param font_name: The font name to sanitize.
    :returns: A valid PostScript font name.
    """
    return re.sub(r"[{}\[\]() <>%/]", "_", font_name)[:63].encode("ascii", "ignore").decode("ascii")


def expand_path(path: str) -> str:
    """
    Expand ``path`` with any variables or shortcuts, such as ``~``, ``$HOME``, etc.

    >>> mcfonts.utils.expand_path("~/Documents")
    '/home/me/Documents'

    :param path: The unexpanded path.
    :returns: The expanded absolute path.
    """
    return os.path.abspath(os.path.expandvars(os.path.expanduser(path)))


def expand_resource_location(file_path: str) -> str:
    """
    Take any path to a texture file and return the Minecraft file string for it.

    :param file_path: A fully-expanded path to any file.
    :returns: A string in the format of "<namespace>?:<dir/>?<file>".
    """
    splitted = re.split(f"[{os.path.sep}:]", file_path)
    try:
        assets = splitted.index("assets")
    except ValueError:
        assets = -1
    if (namespace := splitted[0]) == "minecraft":
        return f"{splitted[assets + 3]}/{'/'.join(splitted[assets + 4:])}"
    return f"{namespace}/{splitted[assets + 3]}/{'/'.join(splitted[assets + 4:])}"


def fit_chars_into_charlist(chars: list[str], charlist_length: int = 16) -> typing.Iterator[str]:
    """
    Given a list of chars, fit them into a charlist whose width is equal to `charlist_length`.

    >>> list(mcfonts.utils.fit_chars_into_charlist(["thisisareallylongcharlist"]))
    ['thisisareallylon', 'gcharlist']
    >>> list(mcfonts.utils.fit_chars_into_charlist(["thisisareallylongcharlist"], 5))
    ['thisi', 'sarea', 'llylo', 'ngcha', 'rlist']

    :param chars: A list of chars.
    :param charlist_length: The width to make each line of characters in the charlist equal to.
    :returns: Yield of lines in a charlist.
    """
    chars = list(mcfonts.utils.charlist_to_chars(chars))
    for i in range(0, len(chars), charlist_length):
        yield "".join(char for char in chars[i : i + charlist_length])


def charlist_to_chars(charlist: list[str]) -> typing.Iterator[str]:
    """
    Given `charlist`, yield all the characters that charlist covers.

    >>> chars = ['agtg', 'b', '5', 'c', 'd', 'e', 'f', '0', '1', '2', '3', '4', '5', '6', '7', '8']
    >>> list(mcfonts.utils.charlist_to_chars(chars)
    ['a', 'g', 't', 'g', 'b', '5', 'c', 'd', 'e', 'f', '0', '1', '2', '3', '4', '5', '6', '7', '8']

    :param charlist: The charlist.
    :returns: A yield of characters.
    """
    for charline in charlist:
        # iterating over lines, charline is a string
        for character in charline:
            # iterating over chars, char is a str[1]
            yield character


def is_image_empty(image: PIL.Image.Image) -> bool:
    """
    Determine if `image` has any pixel data.

    :param image: A :class:`PIL.Image.Image`.
    :returns: If `image` has pixel data.
    """
    extrema = image.getextrema()
    if isinstance(extrema[0], int):
        return extrema[0] == 0 and extrema[1] == 0
    for band in extrema:
        if band[0] != 0 or band[1] != 0:
            return False
    return True


def is_image_invisible(image: PIL.Image.Image) -> bool:
    """
    Determine if `image` has all invisible pixels; if alpha is 0.

    :param image: A :class:`PIL.Image.Image`.
    :returns: If `image` doesn't have any full-alpha pixels.
    """
    if image.mode == "RGBA":
        return all(pixel == 0 for pixel in image.getdata(3))
    if image.mode == "LA":
        return all(pixel == 0 for pixel in image.getdata(1))
    return False


def print_coverage(coverage_report: dict):
    """
    Given a coverage report generated from a :class:`~mcfonts.MinecraftFont`,
    print out a summary of the coverage report.
    This includes percentages of how much of a block is covered by a font.

    :param coverage_report:
        A dictionary of the coverage.
        See :meth:`mcfonts.MinecraftFont.build_coverage_report`.
    """
    print(f"Blocks covered:     {len(coverage_report['blocks']):,}")
    print(f"Characters covered: {coverage_report['chars']:,}")
    for block_name, block_coverage in coverage_report["blocks"].items():
        block_info = tinyunicodeblock.BLOCKS_BYNAME[block_name]
        percentage = (block_coverage / (block_info[1] - block_info[0] + 1)) * 100
        print(
            f"{chr(block_info[0] + 5)} {(f'{block_name}:'.ljust(40))} {block_coverage:,}/"
            f"{(block_info[1] - block_info[0] + 1):,} ({percentage:.2f}%)"
        )


def charlist_from_unicode_range(start: str, end: str, width: int = 16) -> list[str]:
    r"""
    Given a starting character `start`, and an ending character `end`,
    return a charlist that contains these characters in order,
    and whose width is equal to `width`.

    >>> mcfonts.utils.charlist_from_unicode_range("\u2600", "\u26ff")
    [
        '☀☁☂☃☄★☆☇☈☉☊☋☌☍☎☏',
        '☐☑☒☓☔☕☖☗☘☙☚☛☜☝☞☟',
        '☠☡☢☣☤☥☦☧☨☩☪☫☬☭☮☯',
        '☰☱☲☳☴☵☶☷☸☹☺☻☼☽☾☿',
        '♀♁♂♃♄♅♆♇♈♉♊♋♌♍♎♏',
        '♐♑♒♓♔♕♖♗♘♙♚♛♜♝♞♟',
        '♠♡♢♣♤♥♦♧♨♩♪♫♬♭♮♯',
        '♰♱♲♳♴♵♶♷♸♹♺♻♼♽♾♿',
        '⚀⚁⚂⚃⚄⚅⚆⚇⚈⚉⚊⚋⚌⚍⚎⚏',
        '⚐⚑⚒⚓⚔⚕⚖⚗⚘⚙⚚⚛⚜⚝⚞⚟',
        '⚠⚡⚢⚣⚤⚥⚦⚧⚨⚩⚪⚫⚬⚭⚮⚯',
        '⚰⚱⚲⚳⚴⚵⚶⚷⚸⚹⚺⚻⚼⚽⚾⚿',
        '⛀⛁⛂⛃⛄⛅⛆⛇⛈⛉⛊⛋⛌⛍⛎⛏',
        '⛐⛑⛒⛓⛔⛕⛖⛗⛘⛙⛚⛛⛜⛝⛞⛟',
        '⛠⛡⛢⛣⛤⛥⛦⛧⛨⛩⛪⛫⛬⛭⛮⛯',
        '⛰⛱⛲⛳⛴⛵⛶⛷⛸⛹⛺⛻⛼⛽⛾⛿'
    ]
    >>> mcfonts.utils.charlist_from_unicode_range(" ", "\xff", 8)
    [
        ' !"#$%&'',
        '()*+,-./',
        '01234567',
        '89:;<=>?',
        '@ABCDEFG',
        'HIJKLMNO',
        'PQRSTUVW',
        'XYZ[\\]^_',
        '`abcdefg',
        'hijklmno',
        'pqrstuvw',
        'xyz{|}~\x7f',
        '\x80\x81\x82\x83\x84\x85\x86\x87',
        '\x88\x89\x8a\x8b\x8c\x8d\x8e\x8f',
        '\x90\x91\x92\x93\x94\x95\x96\x97',
        '\x98\x99\x9a\x9b\x9c\x9d\x9e\x9f',
        '\xa0¡¢£¤¥¦§',
        '¨©ª«¬\xad®¯',
        '°±²³´µ¶·',
        '¸¹º»¼½¾¿',
        'ÀÁÂÃÄÅÆÇ',
        'ÈÉÊËÌÍÎÏ',
        'ÐÑÒÓÔÕÖ×',
        'ØÙÚÛÜÝÞß',
        'àáâãäåæç',
        'èéêëìíîï',
        'ðñòóôõö÷',
        'øùúûüýþÿ'
    ]

    :param start: The starting single character.
    :param end:
        The ending single character.
        Must be greater than `start`.
    :param width: The number of characters to put in one row of the charlist.
    :returns: A charlist, each string's length equal to `width`.
    """
    start_codepoint = ord(start)
    end_codepoint = ord(end)
    if end_codepoint <= start_codepoint:
        return []
    return list(fit_chars_into_charlist([chr(c) for c in range(start_codepoint, end_codepoint + 1)], width))


def is_charline_empty(charline: str) -> bool:
    r"""
    Given `charline`, return if it contains only spaces or null bytes.

    >>> mcfonts.utils.is_charline_empty("\0\0\0\x20\x20\x20\0\0")
    True

    >>> mcfonts.utils.is_charline_empty("         ")
    True

    >>> mcfonts.utils.is_charline_empty("      xxx")
    False

    :param charline: A single string, likely part_notation of a wider charlist.
    :returns: If `charline` is all spaces or null bytes.
    """
    return all(x in mcfonts.constants.PADDING_CHARS for x in charline)


def color_number(number: float | int) -> str:
    """
    Given `number`, return a colorized and pretty-print version of that number.

    .. note::
        If :data:`mcfonts.colors.USE_COLORS` is False, color will not be applied.

    If `number` is negative, it will be in red.
    If `number` is positive, it will be in green.
    If `number` is zero, it will have no colors.

    :param number: The number, positive or negative.
    :returns: A string representing `number` with color codes.
    """
    if number < 0:
        if mcfonts.colors.USE_COLORS:
            return f"{mcfonts.colors.RED_FORE}{number:,}{mcfonts.colors.RESET_FORE}"
        return f"{number:,}"
    if number == 0:
        return f"{number:,}"
    if mcfonts.colors.USE_COLORS:
        return f"{mcfonts.colors.GREEN_FORE}+{number:,}{mcfonts.colors.RESET_FORE}"
    return f"+{number:,}"



#!/usr/bin/env python3
# coding=utf-8
# SPDX-License-Identifier: MIT
"""
Utilities for the "legacy_unicode" font provider.
Most of these are for either handling codepages, or for building template chars.
"""

import os

import PIL.Image
import mcfonts.constants
import mcfonts.utils.bitmap


def build_template_for_char(char: str, path_to_template_textures: str | None = None) -> PIL.Image.Image:
    """
    Build an image of a template :term:`codepoint`.
    Templates are simple white boxes with a codepoint :term:`texture` inside them.
    This is intended for use with a "legacy_unicode" :term:`provider`.

    :param char:
        A single character.
        This character's codepoint is what goes inside the box.
    :param path_to_template_textures:
        The path to the folder that contains the digit textures.
        See the documentation for :func:`mcfonts.templates.generate_unicode_template_font`.
    :returns: A :class:`PIL.Image.Image` of the char template
    """
    codepoint = ord(char)
    templates = generate_unicode_template_font(path_to_template_textures)
    if codepoint > 0xFFFF:
        box = templates[-6]
        # 0xABCDEF -> A
        box.paste(templates[(codepoint & 0xF00000) >> 20], (2, 2))
        # 0xABCDEF -> B
        box.paste(templates[(codepoint & 0xF0000) >> 16], (6, 2))
        # 0xABCDEF -> C
        box.paste(templates[(codepoint & 0xF000) >> 12], (10, 2))
        # 0xABCDEF -> D
        box.paste(templates[(codepoint & 0xF00) >> 8], (2, 9))
        # 0xABCDEF -> E
        box.paste(templates[(codepoint & 0xF0) >> 4], (6, 9))
        # 0xABCDEF -> F
        box.paste(templates[(codepoint & 0xF)], (10, 9))
    else:
        box = templates[-4]
        # 0xABCD -> A
        box.paste(templates[(codepoint & 0xF000) >> 12], (2, 2))
        # 0xABCD -> B
        box.paste(templates[(codepoint & 0xF00) >> 8], (6, 2))
        # 0xABCD -> C
        box.paste(templates[(codepoint & 0xF0) >> 4], (2, 9))
        # 0xABCD -> D
        box.paste(templates[codepoint & 0xF], (6, 9))
    return box


def generate_unicode_template_font(path: str | None = None) -> dict[int, PIL.Image.Image]:
    """
    Generate a dictionary of hex -> sheet.

    Path must point to a folder where these files exist (PNG format):

    * <0-9>.png
    * <a-f>.png
    * box4.png
    * box6.png

    The file names **are** case-sensitive
    and the textures for hex charlist **must** be 3 pixels wide by 6 pixels tall.

    The sheet for "box6.png" must be 16 x 16 pixels.
    The sheet for "box4.png" must be 12 x 16 pixels.
    Hex charlist are overlayed atop the box.

    Box4 is for chars below 0x10000 (4 points, F4E -> 0F4E; 4).
    Box6 is for chars above 0xFFFF (6 points, 1FAE6 -> 01FAE6; 6).

    Color does not matter.

    This is used for generating Unicode template characters,
    see :func:`mcfonts.utils.legacy_unicode.build_template_for_char`.

    :param path:
        Path to the folder where these textures exist.
        If this is falsy (None, ""), the path will be set automatically from the module and will
        use pre-made character textures that were distributed with this module.
    :returns:
        A dictionary that maps 0-15 to a glyph.
        -4 and -6 are box4 and box6, respectively.
    """
    if path:
        templates = mcfonts.utils.expand_path(path)
    else:
        templates = mcfonts.utils.expand_path(os.path.join(__file__, "../../template_hexchar_textures"))
    return {
        0: PIL.Image.open(f"{templates}/0.png"),
        1: PIL.Image.open(f"{templates}/1.png"),
        2: PIL.Image.open(f"{templates}/2.png"),
        3: PIL.Image.open(f"{templates}/3.png"),
        4: PIL.Image.open(f"{templates}/4.png"),
        5: PIL.Image.open(f"{templates}/5.png"),
        6: PIL.Image.open(f"{templates}/6.png"),
        7: PIL.Image.open(f"{templates}/7.png"),
        8: PIL.Image.open(f"{templates}/8.png"),
        9: PIL.Image.open(f"{templates}/9.png"),
        10: PIL.Image.open(f"{templates}/a.png"),
        11: PIL.Image.open(f"{templates}/b.png"),
        12: PIL.Image.open(f"{templates}/c.png"),
        13: PIL.Image.open(f"{templates}/d.png"),
        14: PIL.Image.open(f"{templates}/e.png"),
        15: PIL.Image.open(f"{templates}/f.png"),
        -4: PIL.Image.open(f"{templates}/box4.png"),
        -6: PIL.Image.open(f"{templates}/box6.png"),
    }


def align_unicode_page(sheet: PIL.Image.Image) -> PIL.Image.Image:
    """
    Align a Unicode page font sheet's characters to the left.

    This function is really a shortcut for
    ``mcfonts.utils.bitmap.align_font_texture(sheet, (16, 16))``.

    :param sheet: The font sheet, not the individual character.
    :returns: The new font sheet.
    """
    return mcfonts.utils.bitmap.align_font_texture(sheet, (16, 16))
